import os
import glob
from pathlib import Path
from vyperdatum.transformer import Transformer
from vyperdatum.utils.raster_utils import raster_metadata, update_raster_wkt
from vyperdatum.utils.vdatum_rest_utils import vdatum_cross_validate
from vyperdatum.utils.crs_utils import pipeline_string
import pyproj as pp
from osgeo import gdal
import re


def get_tiff_files(parent_dir: str, extention: str) -> list:
    tiff_files = []
    for (dirpath, dirnames, filenames) in os.walk(parent_dir):
        for filename in filenames:
            if filename.endswith(extention):
                tiff_files.append(os.sep.join([dirpath, filename]))
    return tiff_files


def transform_with_vyperdatum(input_file, crs_from, crs_to):
    # steps = [
    #             {"crs_from": "EPSG:6347", "crs_to": "EPSG:6318", "v_shift": False},
    #             {"crs_from": "EPSG:6319", "crs_to": "EPSG:6318+NOAA:98", "v_shift": True},
    #             {"crs_from": "EPSG:6318+NOAA:98", "crs_to": "EPSG:6347+NOAA:98", "v_shift": False}
    #         ]

    tf = Transformer(crs_from=crs_from,
                     crs_to=crs_to,
                    #  steps=steps
                     )
    output_file = input_file.replace("Original", "Manual")
    tf.transform_raster(input_file=input_file,
                        output_file=output_file,
                        overview=False,
                        pre_post_checks=True,
                        vdatum_check=True
                        )
    return


def steps_to_concat_pipe(steps):
    concat_pipe = "+proj=pipeline "
    for step in steps:
        pipe = pipeline_string(step["crs_from"], step["crs_to"])
        concat_pipe = f"{concat_pipe} {pipe.split('+proj=pipeline')[1]}"
    return concat_pipe


def transform_with_concat_steps(input_file, output_file):
    steps = [
                {"crs_from": "EPSG:6347", "crs_to": "EPSG:6318", "v_shift": False},
                {"crs_from": "EPSG:6319", "crs_to": "EPSG:6318+NOAA:98", "v_shift": True},
                {"crs_from": "EPSG:6318", "crs_to": "EPSG:6347", "v_shift": False}
            ]


    gdal_warp_concatenated_transform = steps_to_concat_pipe(steps)
    output_vrt = output_file.replace(".tif", ".vrt")
    # output pixel resolution = input pixel resolution
    with gdal.Open(input_file, gdal.GA_ReadOnly) as input_ds:
        geotransform = input_ds.GetGeoTransform()
        xres, yres = geotransform[1], geotransform[5]

    # create vrt transformation template dataset
    ds = gdal.Warp(output_vrt, input_file, format="vrt", outputType=gdal.gdalconst.GDT_Float32,warpOptions = ["APPLY_VERTICAL_SHIFT=YES", "SAMPLE_GRID=YES", "SAMPLE_STEPS=ALL"], errorThreshold = 0, xRes=xres, yRes=yres, targetAlignedPixels=True, coordinateOperation=gdal_warp_concatenated_transform)
    # remove whitespace formatting used in proj pipeline string and put it in GeoTIFF metadata
    gdal_warp_concatenated_transform = re.sub(r'\s{2,}', ' ', gdal_warp_concatenated_transform).strip()
    ds.SetMetadataItem('TIFFTAG_IMAGEDESCRIPTION', gdal_warp_concatenated_transform)

    # execute the transformation from the vrt dataset to output compressed tif
    output_ds = gdal.Translate(output_file, ds, format="GTiff", outputType=gdal.GDT_Float32, creationOptions=["COMPRESS=DEFLATE", "TILED=YES"])
    output_ds = None    
    return

def transform_with_concat_pipe(input_file, output_file):    
    output_vrt = output_file.replace(".tif", ".vrt")

    #projinfo -s EPSG:6347 -t EPSG:9989 --spatial-test intersects --hide-ballpark -o PROJ
    NAD83_2011_UTM18N_to_ITRF2020 = """
    +step +inv +proj=utm +zone=18 +ellps=GRS80
    +step +proj=cart +ellps=GRS80
    +step +inv +proj=helmert
        +x=1.0039 +y=-1.90961 +z=-0.54117 +rx=0.02678138 +ry=-0.00042027 +rz=0.01093206 +s=-5.109e-05
        +dx=0.00079 +dy=-0.0007 +dz=-0.00124 +drx=6.667e-05 +dry=-0.00075744 +drz=-5.133e-05 +ds=-7.201e-05
        +t_epoch=2010 +convention=coordinate_frame
    +step +inv +proj=cart +ellps=GRS80"""

    #projinfo -s EPSG:9990 -t EPSG:6347 --spatial-test intersects --hide-ballpark -o PROJ
    ITRF2020_2010_to_NAD83_2011_18N = """
    +step +proj=cart +ellps=GRS80
    +step +proj=helmert +x=1.0039 +y=-1.90961 +z=-0.54117 +rx=0.02678138
            +ry=-0.00042027 +rz=0.01093206 +s=-5.109e-05 +dx=0.00079 +dy=-0.0007
            +dz=-0.00124 +drx=6.667e-05 +dry=-0.00075744 +drz=-5.133e-05 +ds=-7.201e-05
            +t_epoch=2010 +convention=coordinate_frame
    +step +inv +proj=cart +ellps=GRS80
    +step +proj=utm +zone=18 +ellps=GRS80"""

    MLLW_ITRF2020_2020 = "us_noaa_nos_MLLW-ITRF2020_2020.0_(nwldatum_4.7.0_20240621).tif"

    gdal_warp_concatenated_transform = f"""\
    +proj=pipeline
    {NAD83_2011_UTM18N_to_ITRF2020}
    +step +inv +proj=vgridshift +grids={MLLW_ITRF2020_2020} +multiplier=1
    +step +proj=push +v_3
    {ITRF2020_2010_to_NAD83_2011_18N}
    +step +proj=pop +v_3"""

    # output pixel resolution = input pixel resolution
    with gdal.Open(input_file, gdal.GA_ReadOnly) as input_ds:
        geotransform = input_ds.GetGeoTransform()
        xres, yres = geotransform[1], geotransform[5]

    # create vrt transformation template dataset
    ds = gdal.Warp(output_vrt, input_file, format="vrt", outputType=gdal.gdalconst.GDT_Float32,warpOptions = ["APPLY_VERTICAL_SHIFT=YES", "SAMPLE_GRID=YES", "SAMPLE_STEPS=ALL"], errorThreshold = 0, xRes=xres, yRes=yres, targetAlignedPixels=True, coordinateOperation=gdal_warp_concatenated_transform)
    # remove whitespace formatting used in proj pipeline string and put it in GeoTIFF metadata
    gdal_warp_concatenated_transform = re.sub(r'\s{2,}', ' ', gdal_warp_concatenated_transform).strip()
    ds.SetMetadataItem('TIFFTAG_IMAGEDESCRIPTION', gdal_warp_concatenated_transform)

    # execute the transformation from the vrt dataset to output compressed tif
    output_ds = gdal.Translate(output_file, ds, format="GTiff", outputType=gdal.GDT_Float32, creationOptions=["COMPRESS=DEFLATE", "TILED=YES"])
    output_ds = None    
    return

def nad83_transform_with_concat_pipe(input_file, output_file):    
    output_vrt = output_file.replace(".tif", ".vrt")

    #projinfo -s EPSG:6347 -t EPSG:6319 --spatial-test intersects --hide-ballpark -o PROJ
    NAD83_2011_UTM18N_to_GEOGRAPHIC = "+step +inv +proj=utm +zone=18 +ellps=GRS80"
    GEOGRAPHIC_to_NAD83_2011_UTM18N = "+step +proj=utm +zone=18 +ellps=GRS80"
    # NAD83_2011_UTM18N_to_GEOGRAPHIC = """
    # +step +inv +proj=utm +zone=18 +ellps=GRS80
    # +step +proj=unitconvert +xy_in=rad +z_in=m +xy_out=deg +z_out=m
    # +step +proj=axisswap +order=2,1
    # """
    # GEOGRAPHIC_to_NAD83_2011_UTM18N = """
    # +step +proj=axisswap +order=2,1
    # +step +proj=unitconvert +xy_in=deg +xy_out=rad
    # +step +proj=utm +zone=18 +ellps=GRS80
    # """

    horz_push = "+step +proj=push +v_1 +v_2"
    horz_pop = "+step +proj=pop +v_1 +v_2"

    #projinfo -s EPSG:6319 -t EPSG:6318+NOAA:98 --spatial-test intersects --hide-ballpark -o PROJ
    NAD83_2011_to_NWLD = "+step +inv +proj=vgridshift +grids=us_noaa_nos_MLLW-NAD83(2011)_2010.0_(nwldatum_4.7.0_20240621).tif +multiplier=1"
    # NAD83_2011_to_NWLD = """
    # +step +proj=axisswap +order=2,1
    # +step +proj=unitconvert +xy_in=deg +xy_out=rad
    # +step +inv +proj=vgridshift
    #         +grids=us_noaa_nos_MLLW-NAD83(2011)_2010.0_(nwldatum_4.7.0_20240621).tif
    #         +multiplier=1
    # +step +proj=unitconvert +xy_in=rad +xy_out=deg
    # +step +proj=axisswap +order=2,1
    # """
    
    gdal_warp_concatenated_transform = f"""\
    +proj=pipeline
    {NAD83_2011_UTM18N_to_GEOGRAPHIC}

    {NAD83_2011_to_NWLD}

    {GEOGRAPHIC_to_NAD83_2011_UTM18N}
    """

    # output pixel resolution = input pixel resolution
    with gdal.Open(input_file, gdal.GA_ReadOnly) as input_ds:
        geotransform = input_ds.GetGeoTransform()
        xres, yres = geotransform[1], geotransform[5]

    # create vrt transformation template dataset
    ds = gdal.Warp(output_vrt, input_file, format="vrt", outputType=gdal.gdalconst.GDT_Float32,warpOptions = ["APPLY_VERTICAL_SHIFT=YES", "SAMPLE_GRID=YES", "SAMPLE_STEPS=ALL"], errorThreshold = 0, xRes=xres, yRes=yres, targetAlignedPixels=True, coordinateOperation=gdal_warp_concatenated_transform)
    # remove whitespace formatting used in proj pipeline string and put it in GeoTIFF metadata
    gdal_warp_concatenated_transform = re.sub(r'\s{2,}', ' ', gdal_warp_concatenated_transform).strip()
    ds.SetMetadataItem('TIFFTAG_IMAGEDESCRIPTION', gdal_warp_concatenated_transform)

    # execute the transformation from the vrt dataset to output compressed tif
    output_ds = gdal.Translate(output_file, ds, format="GTiff", outputType=gdal.GDT_Float32, creationOptions=["COMPRESS=DEFLATE", "TILED=YES"])
    output_ds = None    
    return




if __name__ == "__main__":
    files = glob.glob(r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\NC\Original\**\*.tif", recursive=True)
    files = glob.glob(r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\RSD\MD\Original\**\*.tif", recursive=True)
    files = [r"C:\Users\mohammad.ashkezari\Desktop\test\Original\clip_NC.tif"]
    files = glob.glob(r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\RSD\VA\Original\*.tif", recursive=True)



    crs_from = "EPSG:6347"
    crs_to = "EPSG:6347+NOAA:98"

    # files = [r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\NC\Original\VA1803-TB-C_BLK-01\VA1803-TB-C_BLK-01_US4VA1FD_ellipsoidal_dem_b1.tif"]

    for i, input_file in enumerate(files[:]):
        print(f"{i+1}/{len(files)}: {input_file}")
        transform_with_vyperdatum(input_file, crs_from, crs_to)

        ###########################
        # try:
        #     output_file = input_file.replace("Original", "Manual")
        #     Path(os.path.split(output_file)[0]).mkdir(parents=True, exist_ok=True)
        #     # transform_with_concat_pipe(input_file, output_file)
        #     # nad83_transform_with_concat_pipe(input_file, output_file)
        #     transform_with_concat_steps(input_file, output_file)

        #     update_raster_wkt(output_file, pp.CRS(crs_to).to_wkt())
        #     input_metadata, output_metadata = raster_metadata(input_file), raster_metadata(output_file)
        #     passed, cross_df = vdatum_cross_validate(s_wkt=input_metadata["wkt"],
        #                                              t_wkt=output_metadata["wkt"],
        #                                              n_sample=20,
        #                                              s_raster_metadata=input_metadata,
        #                                              t_raster_metadata=output_metadata,
        #                                              s_point_samples=None,
        #                                              t_point_samples=None,
        #                                              tolerance=0.3,
        #                                              raster_sampling_band=1,
        #                                              region=None,
        #                                              pivot_h_crs="EPSG:6318",
        #                                              s_h_frame=None,
        #                                              s_v_frame=None,
        #                                              s_h_zone=None,
        #                                              t_h_frame=None,
        #                                              t_v_frame=None,
        #                                              t_h_zone=None
        #                                              )

        #     # if not passed:
        #     cross_df.to_csv(Path(output_file).parent.absolute()/Path(f"{os.path.split(input_file)[1]}_vdatum_check.csv"), index=False)
        # except Exception as e:
        #         efile = open(Path(output_file).parent.absolute()/Path(f"{os.path.split(input_file)[1]}_error.txt"), "w")
        #         efile.write(str(e))
        #         efile.close()

        ###########################

        print(f'\n{"*"*50} {i+1}/{len(files)} Completed {"*"*50}\n')



import os
import pathlib
import shutil
from importlib.metadata import version
import io
import sys
import logging
from pathlib import Path
import re
import json
from typing import Union, Optional
from colorama import Fore, Style
import pyproj as pp
from pyproj._transformer import AreaOfInterest
import numpy as np
from osgeo import gdal, osr, ogr
from tqdm import tqdm
from vyperdatum.utils import raster_utils, crs_utils, drivers_utils
from vyperdatum.utils.raster_utils import (raster_metadata,
                                           update_raster_wkt,
                                           overwrite_with_original,
                                           apply_nbs_band_standards,
                                           create_cutline_from_grid)
from vyperdatum.utils.vdatum_rest_utils import vdatum_cross_validate
from vyperdatum.drivers import vrbag, laz, npz, pdal_based, gparq, xyz
from vyperdatum.pipeline import nwld_ITRF2020_steps, nwld_NAD832011_steps

logger = logging.getLogger("root_logger")
gdal.UseExceptions()

# os.environ["CPL_DEBUG"] = "ON"
# os.environ["CPL_LOG_ERRORS"] = "ON"
# os.environ["PROJ_DEBUG"] = "3"


class Transformer():
    def __init__(self,
                 crs_from: Union[pp.CRS, int, str],
                 crs_to: Union[pp.CRS, int, str],
                 steps: Optional[list[str]] = None
                 ) -> None:
        """

        Raises
        ----------
        ValueError
            If the transformation steps cannot be validated.

        Parameters
        ----------
        crs_from: pyproj.crs.CRS or input used to create one
            Projection of input data.
        crs_to: pyproj.crs.CRS or input used to create one
            Projection of output data.
        steps: Optional[list[dict]]
            A list of dicts containing source and target CRSs in form of `authority:code`,
            and a boolean key to signify if the step impose a vertical shift. This parameter
            represents the overall transformation steps connecting the `crs_from` to `crs_to`.
            When `None` is passed, vyperdatum will attempt to automatically determine the steps
            from `crs_from` to `crs_to`.
            Example:
            steps = [{"crs_from": "EPSG:6346", "crs_to": "EPSG:6318", "v_shift": False},
                     {"crs_from": "EPSG:6319", "crs_to": "EPSG:6318+NOAA:98", "v_shift": True},
                     {"crs_from": "EPSG:6318", "crs_to": "EPSG:6346", "v_shift": False}
                    ]
        """

        if not isinstance(crs_from, pp.CRS):
            crs_from = pp.CRS(crs_from)
        if not isinstance(crs_to, pp.CRS):
            crs_to = pp.CRS(crs_to)
        self.crs_from = crs_from
        self.crs_to = crs_to
        self.steps = steps
        if not self.steps:
            # self.steps = [crs_utils.auth_code(self.crs_from), crs_utils.auth_code(self.crs_to)]
            h0, v0 = crs_utils.crs_components(self.crs_from)
            h1, v1 = crs_utils.crs_components(self.crs_to)
            # self.steps = nwld_ITRF2020_steps(h0, v0, h1, v1)
            self.steps = nwld_NAD832011_steps(h0, v0, h1, v1)
        if not crs_utils.validate_transform_steps_dict(self.steps):
            raise ValueError(f"Invalid transformation pipeline: {self.steps}.")
        return

    @classmethod
    def from_GTiff_raster(cls,
                          input_file: str,
                          crs_to: Union[pp.CRS, int, str],
                          steps: Optional[list[dict]]) -> "Transformer":
        """
        Create a Transformer instance from a GeoTiff raster file.

        Raises
        ----------
        FileNotFoundError
            If the input file is not found.
        ValueError
            If the input raster does not have the `Vyperdatum_Metadata` metadata tag.

        Parameters
        ----------
        input_file: str
            Path to the input raster file.
        crs_to: pyproj.crs.CRS or input used to create one
            Projection of output data.
        steps: Optional[list[dict]]
            A list of dicts containing source and target CRSs in form of `authority:code`,
            and a boolean key to signify if the step impose a vertical shift. This parameter
            represents the overall transformation steps connecting the `crs_from` to `crs_to`.
            When `None` is passed, vyperdatum will attempt to automatically determine the steps.        
        """
        if not os.path.isfile(input_file):
            raise FileNotFoundError(f"The input file not found at {input_file}.")
        meta = raster_utils.raster_metadata(input_file)
        if "Vyperdatum_Metadata" not in meta:
            raise ValueError("The input raster file does not have the `Vyperdatum_Metadata` tag.")
        vyperdatum_metadata = json.loads(meta["Vyperdatum_Metadata"])
        crs_from = pp.CRS(vyperdatum_metadata["wkt"])
        return cls(crs_from=crs_from, crs_to=crs_to, steps=steps)

    @staticmethod
    def gdal_extensions() -> list[str]:
        """
        Return a lower-cased list of driver names supported by gdal.

        Returns
        -------
        list[str]
        """
        return sorted(
            ["." + gdal.GetDriver(i).ShortName.lower() for i in range(gdal.GetDriverCount())]
            + [".tif", ".tiff"]
            )

    def _validate_input_file(self, input_file: str) -> bool:
        """
        Check if the input file (`input_file`) exists and supported by GDAL.

        Raises
        -------
        FileNotFoundError:
            If the input raster file is not found.
        NotImplementedError:
            If the input vector file is not supported by gdal.

        Parameters
        -----------
        input_file: str
            Path to the input raster file (gdal supported).

        Returns
        -----------
        bool
            True if passes all checks, otherwise False.
        """
        passed = False
        if "vsimem" not in [s.lower() for s in input_file.split("/")] and not os.path.isfile(input_file):
            raise FileNotFoundError(f"The input raster file not found at {input_file}.")
        if pathlib.Path(input_file).suffix.lower() not in self.gdal_extensions():
            raise NotImplementedError(f"{pathlib.Path(input_file).suffix} is not supported")
        passed = True
        return passed

    def transform(self,
                  input_file: str,
                  output_file: str,
                  pre_post_checks: bool = True,
                  vdatum_check: bool = False,
                  **kwargs
                  ) -> bool:
        """
        Top-level transform method.

        Parameters
        -----------
        input_file: str
            Path to the input file.
        output_file: str
            Path to the output transformed file.
        pre_post_checks: bool, default=True
            If True, runs a series of validation checks, such as validating the input and output
            CRSs, before and after transformation operation.

        Raises
        -------
        FileNotFoundError:
            If the input  file is not found.
        NotImplementedError:
            If the input file is not supported by vyperdatum.

        Returns
        -----------
        bool:
            True if successful, otherwise False.
        """
        try:
            success = False    
            if not os.path.isfile(input_file):
                raise FileNotFoundError(f"The input file not found at {input_file}.")

            if vrbag.is_vr(fname=input_file):
                logger.info(f"Identified as vrbag file: {input_file}")
                success = self.transform_vrbag(input_file=input_file,
                                               output_file=output_file,
                                               pre_post_checks=pre_post_checks,
                                               vdatum_check=vdatum_check
                                               )
            elif gparq.GeoParquet(input_file=input_file, invalid_error=False).is_valid:
                logger.info(f"Identified as geoparquet file: {input_file}")
                success = self.transform_geoparquet(input_file=input_file,
                                                    output_file=output_file,
                                                    pre_post_checks=pre_post_checks,
                                                    vdatum_check=vdatum_check
                                                    )
            elif laz.LAZ(input_file=input_file, invalid_error=False).is_valid:
                logger.info(f"Identified as laz file: {input_file}")
                success = self.transform_laz(input_file=input_file,
                                             output_file=output_file,
                                             pre_post_checks=pre_post_checks,
                                             vdatum_check=vdatum_check
                                             )
            elif npz.NPZ(input_file=input_file, invalid_error=False).is_valid:
                logger.info(f"Identified as npz file: {input_file}")
                success = self.transform_npz(input_file=input_file,
                                             output_file=output_file,
                                             pre_post_checks=pre_post_checks,
                                             vdatum_check=vdatum_check
                                             )
            elif xyz.XYZ(input_file=input_file, invalid_error=False).is_valid:
                logger.info(f"Identified as xyz file: {input_file}")
                success = self.transform_xyz(input_file=input_file,
                                             output_file=output_file,
                                             pre_post_checks=pre_post_checks,
                                             vdatum_check=vdatum_check,
                                             **kwargs
                                             )
            elif pathlib.Path(input_file).suffix.lower() in self.gdal_extensions():
                logger.info(f"Identified as GDAL-supported raster file: {input_file}")
                success = self.transform_raster(input_file=input_file,
                                                output_file=output_file,
                                                pre_post_checks=pre_post_checks,
                                                vdatum_check=vdatum_check
                                                )
            elif pdal_based.PDAL(input_file=input_file,
                                 output_file=output_file, invalid_error=False).is_valid:
                logger.info(f"Identified as PDAL-supported file: {input_file}")
                success = self.transform_pdal(input_file=input_file,
                                              output_file=output_file,
                                              pre_post_checks=pre_post_checks,
                                              vdatum_check=vdatum_check
                                              )
            # elif vector files
            else:
                raise NotImplementedError(f"Unsupported input file: {input_file}")
        finally:
            return success

    def transform_points(self,
                         x: Union[list, np.ndarray],
                         y: Union[list, np.ndarray],
                         z: Union[list, np.ndarray],
                         always_xy: bool = False,
                         vdatum_check: bool = False,
                         area_of_interest: Optional[AreaOfInterest] = None,
                         authority: Optional[str] = None,
                         accuracy: Optional[float] = None,
                         allow_ballpark: Optional[bool] = True,
                         force_over: bool = False,
                         only_best: Optional[bool] = True
                         ) -> tuple[Optional[Union[list, np.ndarray]],
                                    Optional[Union[list, np.ndarray]],
                                    Optional[Union[list, np.ndarray]]]:
        """
        Conduct point transformation between two coordinate reference systems.        

        Parameters
        ----------
        x: numeric array
           Input x coordinate(s).
        y: numeric array
           Input y coordinate(s).
        z: numeric array, optional
           Input z coordinate(s).
        always_xy: bool, default=False
            If true, the transform method will accept as input and return as output
            coordinates using the traditional GIS order, that is longitude, latitude
            for geographic CRS and easting, northing for most projected CRS.
        vdatum_check: bool, default=False
            If True, a random sample of the transformed data are compared with transformation
            outcomes produced by Vdatum REST API.
        area_of_interest: :class:`.AreaOfInterest`, optional
            The area of interest to help select the transformation.
        authority: str, optional
            When not specified, coordinate operations from any authority will be
            searched, with the restrictions set in the
            authority_to_authority_preference database table related to the
            authority of the source/target CRS themselves. If authority is set
            to “any”, then coordinate operations from any authority will be
            searched. If authority is a non-empty string different from "any",
            then coordinate operations will be searched only in that authority
            namespace (e.g. EPSG).
        accuracy: float, optional
            The minimum desired accuracy (in metres) of the candidate
            coordinate operations.
        allow_ballpark: bool, optional, default=True
            Set to False to disallow the use of Ballpark transformation
            in the candidate coordinate operations. Default is to allow.
        force_over: bool, default=False
            If True, it will to force the +over flag on the transformation.
            Requires PROJ 9+.
        only_best: bool, optional, default=True
            Can be set to True to cause PROJ to error out if the best
            transformation known to PROJ and usable by PROJ if all grids known and
            usable by PROJ were accessible, cannot be used. Best transformation should
            be understood as the transformation returned by
            :c:func:`proj_get_suggested_operation` if all known grids were
            accessible (either locally or through network).
            Note that the default value for this option can be also set with the
            :envvar:`PROJ_ONLY_BEST_DEFAULT` environment variable, or with the
            ``only_best_default`` setting of :ref:`proj-ini`.
            The only_best kwarg overrides the default value if set.
            Requires PROJ 9.2+.

        Returns
        -----------
        bool:
            True if successful, otherwise False.            
        numeric scalar or array:
           Transformed x coordinate(s).
        numeric scalar or array
           Transformed y coordinate(s).
        numeric scalar or array, optional
           Transformed z coordinate(s).
        """

        try:
            success = False
            xt, yt, zt = x.copy(), y.copy(), z.copy()
            for i in range(len(self.steps)):
                logger.info(f"Step {i+1}/{len(self.steps)}:"
                            f" {self.steps[i]['crs_from']} --> {self.steps[i]['crs_to']}")
                xt, yt, zt = pp.Transformer.from_crs(crs_from=pp.CRS(self.steps[i]["crs_from"]),
                                                     crs_to=pp.CRS(self.steps[i]["crs_to"]),
                                                     always_xy=always_xy,
                                                     area_of_interest=area_of_interest,
                                                     authority=authority,
                                                     accuracy=accuracy,
                                                     allow_ballpark=allow_ballpark,
                                                     force_over=force_over,
                                                     only_best=only_best
                                                     ).transform(xt, yt, zt)
            success = True
            if vdatum_check:
                vdatum_cv, vdatum_df = vdatum_cross_validate(s_wkt=pp.CRS(self.steps[0]["crs_from"]).to_wkt(),
                                                             t_wkt=pp.CRS(self.steps[-1]["crs_to"]).to_wkt(),
                                                             n_sample=20,
                                                             s_raster_metadata=None,
                                                             t_raster_metadata=None,
                                                             s_point_samples=list(zip(x, y, z)),
                                                             t_point_samples=list(zip(xt, yt, zt)),
                                                             tolerance=0.3,
                                                             raster_sampling_band=1,
                                                             region=None,
                                                             pivot_h_crs="EPSG:6318",
                                                             s_h_frame=None,
                                                             s_v_frame=None,
                                                             s_h_zone=None,
                                                             t_h_frame=None,
                                                             t_v_frame=None,
                                                             t_h_zone=None
                                                            )
                if not vdatum_cv:
                    success = False
                    csv_path = os.path.join(os.getcwd(), "vdatum_check.csv")
                    vdatum_df.to_csv(csv_path, index=False)
                    logger.info(f"{Fore.RED}Vdatum checks on point data failed. "
                                f"VDatum API outputs stored at: {csv_path}")
                    print(Style.RESET_ALL)
                    return success, None, None, None

        except Exception:
            logger.exception("Error while running the point transformation.")
            return success, None, None, None
        finally:
            return success, xt, yt, zt

    def transform_vrbag(self,
                        input_file: str,
                        output_file: str,
                        pre_post_checks: bool = True,
                        vdatum_check: bool = True
                        ) -> bool:
        """
        Transform variable resolution BAG file.

        Parameters
        -----------
        input_file: str
            Path to the input vrbag file.
        output_file: str
            Path to the output transformed vrbag file.
        pre_post_checks: bool, default=True
            If True, runs a series of validation checks, such as validating the input and output
            CRSs, before and after transformation operation.
        vdatum_check: bool, default=True
            If True, a random sample of the transformed data are compared with transformation
            outcomes produced by Vdatum REST API.

        Raises
        -------
        FileNotFoundError:
            If the input file is not found.
        TypeError
            If the passed BAG file is not a valid variable resolution bag file.

        Returns
        -----------
        bool:
            True if successful, otherwise False.
        """
        try:
            success = False
            if not os.path.isfile(input_file):
                raise FileNotFoundError(f"The input file not found at {input_file}.")
            if not vrbag.is_vr(fname=input_file):
                msg = (f"The following file is not a valid variable resolution bag file: {input_file}")
                logger.exception(msg)
                raise TypeError(msg)
            pathlib.Path(os.path.split(output_file)[0]).mkdir(parents=True, exist_ok=True)
            shutil.copy2(input_file, output_file)
            if pre_post_checks:
                drivers_utils.vrbag_pre_transformation_checks(file_path=input_file,
                                                              source_crs=self.crs_from
                                                              )
            success = vrbag.transform(fname=output_file,
                                      tf=self, point_transformation=True,
                                      vdatum_check=vdatum_check)
            if pre_post_checks:
                drivers_utils.vrbag_post_transformation_checks(file_path=output_file,
                                                               target_crs=self.crs_to
                                                               )
        except Exception as e:
            logger.exception(f"Exception in `transform_vrbag()`: {str(e)}")
            if os.path.isfile(output_file):
                os.remove(output_file)
        finally:
            return success

    def transform_laz(self,
                      input_file: str,
                      output_file: str,
                      pre_post_checks: bool = True,
                      vdatum_check: bool = True
                      ) -> bool:
        """
        Transform point-cloud LAZ file.

        Parameters
        -----------
        input_file: str
            Path to the input laz file.
        output_file: str
            Path to the output transformed laz file.
        pre_post_checks: bool, default=True
            If True, runs a series of validation checks, such as validating the input and output
            CRSs, before and after transformation operation.
        vdatum_check: bool, default=True
            If True, a random sample of the transformed data are compared with transformation
            outcomes produced by Vdatum REST API.

        Raises
        -------
        FileNotFoundError:
            If the input file is not found.
        TypeError
            If the passed LAZ file is not valid.

        Returns
        -----------
        bool:
            True if successful, otherwise False.
        """
        try:
            success = False
            if not os.path.isfile(input_file):
                raise FileNotFoundError(f"The input file not found at {input_file}.")
            pathlib.Path(os.path.split(output_file)[0]).mkdir(parents=True, exist_ok=True)
            shutil.copy2(input_file, output_file)
            lz = laz.LAZ(input_file=output_file)
            if pre_post_checks:
                drivers_utils.laz_pre_transformation_checks(file_path=input_file,
                                                            source_crs=self.crs_from
                                                            )
            success = lz.transform(transformer_instance=self, vdatum_check=vdatum_check)
            if pre_post_checks:
                drivers_utils.laz_post_transformation_checks(file_path=output_file,
                                                             target_crs=self.crs_to
                                                             )
        except Exception as e:
            logger.exception(f"Exception in `transform_laz()`: {str(e)}")
            if os.path.isfile(output_file):
                os.remove(output_file)
        finally:
            return success

    def transform_xyz(self,
                      input_file: str,
                      output_file: str,
                      pre_post_checks: bool = True,
                      vdatum_check: bool = True,
                      **kwargs
                      ) -> bool:
        """
        Transform point-cloud XYZ file.

        Parameters
        -----------
        input_file: str
            Path to the input xyz file.
        output_file: str
            Path to the output transformed xyz file.
        pre_post_checks: bool, default=True
            If True, runs a series of validation checks, such as validating the input and output
            CRSs, before and after transformation operation.
        vdatum_check: bool, default=True
            If True, a random sample of the transformed data are compared with transformation
            outcomes produced by Vdatum REST API.

        Raises
        -------
        FileNotFoundError:
            If the input file is not found.
        TypeError
            If the passed xyz file is not valid.

        Returns
        -----------
        bool:
            True if successful, otherwise False.
        """
        try:
            success = False
            if not os.path.isfile(input_file):
                raise FileNotFoundError(f"The input file not found at {input_file}.")
            pathlib.Path(os.path.split(output_file)[0]).mkdir(parents=True, exist_ok=True)
            xyz_ins = xyz.XYZ(input_file=input_file, **kwargs)

            success = xyz_ins.transform(transformer_instance=self,
                                        output_file=output_file,
                                        pre_post_checks=pre_post_checks,
                                        vdatum_check=vdatum_check)
        except Exception as e:
            logger.exception(f"Exception in `transform_xyz()`: {str(e)}")
            if os.path.isfile(output_file):
                os.remove(output_file)
        finally:
            return success

    def transform_geoparquet(self,
                             input_file: str,
                             output_file: str,
                             pre_post_checks: bool = True,
                             vdatum_check: bool = True
                             ) -> bool:
        """
        Transform a geoparquet point file.

        Parameters
        -----------
        input_file: str
            Path to the input geoparquet file.
        output_file: str
            Path to the output transformed file.
        pre_post_checks: bool, default=True
            If True, runs a series of validation checks, such as validating the input and output
            CRSs, before and after transformation operation.
        vdatum_check: bool, default=True
            If True, a random sample of the transformed data are compared with transformation
            outcomes produced by Vdatum REST API.

        Raises
        -------
        FileNotFoundError:
            If the input file is not found.
        TypeError
            If the passed file is not valid.

        Returns
        -----------
        bool:
            True if successful, otherwise False.
        """
        try:
            success = False
            if not os.path.isfile(input_file):
                raise FileNotFoundError(f"The input file not found at {input_file}.")
            pathlib.Path(os.path.split(output_file)[0]).mkdir(parents=True, exist_ok=True)
            gp = gparq.GeoParquet(input_file=input_file)

            success = gp.transform(transformer_instance=self,
                                   output_file=output_file,
                                   pre_post_checks=pre_post_checks,
                                   vdatum_check=vdatum_check)
        except Exception as e:
            logger.exception(f"Exception in `transform_geoparquet()`: {str(e)}")
            if os.path.isfile(output_file):
                os.remove(output_file)
        finally:
            return success

    def transform_npz(self,
                      input_file: str,
                      output_file: str,
                      pre_post_checks: bool = True,
                      vdatum_check: bool = True
                      ) -> bool:
        """
        Transform a numpy npz file.

        Parameters
        -----------
        input_file: str
            Path to the input npz file.
        output_file: str
            Path to the output transformed npz file.
        pre_post_checks: bool, default=True
            If True, runs a series of validation checks, such as validating the input and output
            CRSs, before and after transformation operation.
        vdatum_check: bool, default=True
            If True, a random sample of the transformed data are compared with transformation
            outcomes produced by Vdatum REST API.

        Raises
        -------
        FileNotFoundError:
            If the input file is not found.
        TypeError
            If the passed npz file is not valid.

        Returns
        -----------
        bool:
            True if successful, otherwise False.
        """
        try:
            success = False
            if not os.path.isfile(input_file):
                raise FileNotFoundError(f"The input file not found at {input_file}.")
            pathlib.Path(os.path.split(output_file)[0]).mkdir(parents=True, exist_ok=True)
            shutil.copy2(input_file, output_file)
            nz = npz.NPZ(input_file=output_file)
            if pre_post_checks:
                drivers_utils.npz_pre_transformation_checks(file_path=input_file,
                                                            source_crs=self.crs_from
                                                            )
            success = nz.transform(transformer_instance=self, vdatum_check=vdatum_check)
            if pre_post_checks:
                drivers_utils.npz_post_transformation_checks(file_path=input_file,
                                                             target_crs=self.crs_to
                                                             )
        except Exception as e:
            logger.exception(f"Exception in `transform_npz()`: {str(e)}")
            if os.path.isfile(output_file):
                os.remove(output_file)
        finally:
            return success

    def transform_pdal(self,
                       input_file: str,
                       output_file: str,
                       pre_post_checks: bool = True,
                       vdatum_check: bool = True
                       ) -> bool:
        """
        Transform point-cloud data using PDAL.

        Parameters
        -----------
        input_file: str
            Path to the input file.
        output_file: str
            Path to the output transformed file.
        pre_post_checks: bool, default=True
            If True, runs a series of validation checks, such as validating the input and output
            CRSs, before and after transformation operation.
        vdatum_check: bool, default=True
            If True, a random sample of the transformed data are compared with transformation
            outcomes produced by Vdatum REST API.

        Raises
        -------
        FileNotFoundError:
            If the input file is not found.
        TypeError
            If the passed file is not valid.

        Returns
        -----------
        bool:
            True if successful, otherwise False.
        """
        # TODO implement vdatum_check in pd.transform()
        try:
            success = False
            if not input_file.lower().startswith("http") and not os.path.isfile(input_file):
                raise FileNotFoundError(f"The input file not found at {input_file}.")
            pathlib.Path(os.path.split(output_file)[0]).mkdir(parents=True, exist_ok=True)
            pdl = pdal_based.PDAL(input_file=input_file, output_file=output_file)
            if pre_post_checks:
                drivers_utils.pdal_pre_transformation_checks(file_path=input_file,
                                                             source_crs=self.crs_from
                                                             )
            success = pdl.transform(transformer_instance=self, vdatum_check=vdatum_check)
            if pre_post_checks:
                drivers_utils.pdal_post_transformation_checks(file_path=input_file,
                                                              target_crs=self.crs_to
                                                              )
        except Exception as e:
            logger.exception(f"Exception in `transform_pdal()`: {e}")
            if os.path.isfile(output_file):
                os.remove(output_file)
        finally:
            return success

    def transform_raster(self,
                         input_file: str,
                         output_file: str,
                         overview: bool = False,
                         pre_post_checks: bool = True,
                         vdatum_check: bool = True
                         ) -> bool:
        """
        Transform the gdal-supported input rater file (`input_file`) and store the
        transformed file on the local disk (`output_file`).

        Raises
        -------
        FileNotFoundError:
            If the input raster file is not found.
        NotImplementedError:
            If the input file is not supported by gdal.

        Parameters
        -----------
        input_file: str
            Path to the input raster file (gdal supported).
        output_file: str
            Path to the transformed raster file.
        overview: bool, default=True
            If True, overview bands are added to the output raster file (only GTiff support).
        pre_post_checks: bool, default=True
            If True, runs a series of validation checks, such as validating the input and output
            CRSs, before and after transformation operation.
        vdatum_check: bool, default=True
            If True, a random sample of the transformed data are compared with transformation
            outcomes produced by Vdatum REST API.


        Returns
        --------
        bool:
            True if successful, otherwise False.
        """
        def steps_to_concat_pipe(steps, input_metadata):
            concat_pipe = "+proj=pipeline "
            v_shift = False
            grid_files = []  # Track grid files used
            for step in steps:
                pipe = crs_utils.pipeline_string(step["crs_from"], step["crs_to"], input_metadata)
                concat_pipe = f"{concat_pipe} {pipe.split('+proj=pipeline')[1]}"
                if step["v_shift"]:
                    v_shift = True
                    # Extract grid file from pipeline string
                    match = re.search(r'\+grids=([^\s]+)', pipe)
                    if match:
                        grid_files.append(match.group(1))
            return concat_pipe, v_shift, grid_files

        self._validate_input_file(input_file)
        try:
            success = False
            pathlib.Path(os.path.split(output_file)[0]).mkdir(parents=True, exist_ok=True)
            input_metadata = raster_metadata(input_file)
            pipe, v_shift, grid_files = steps_to_concat_pipe(self.steps, input_metadata)

            logger.info(f"Transformation Steps: {self.steps}")
            logger.info(f"Concatenated PROJ pipeline:\n{pipe}\n")
            output_vrt = Path(output_file).with_suffix(".vrt")
            with gdal.Open(input_file, gdal.GA_ReadOnly) as input_ds:
                geotransform = input_ds.GetGeoTransform()
                xres, yres = geotransform[1], geotransform[5]



            # Create cutline if vertical shift and NWLD grids are involved, otherwise return None
            # TODO: what if more than one NWLD grids are involved?
            cutline_path = raster_utils.create_cutline_file(v_shift, grid_files,
                                                            cutline_path=str(Path(output_file).parent / f"{Path(output_file).stem}_cutline.gpkg"),
                                                            input_metadata=input_metadata)


            wopt = ["SAMPLE_GRID=YES", "SAMPLE_STEPS=ALL"]
            if v_shift:
                wopt.append("APPLY_VERTICAL_SHIFT=YES")

            if cutline_path:
                # if cutline is created, clip input raster to the underlying grid to avoid gdal/warp failures
                input_file = raster_utils.clip_raster_to_cutline(input_file, cutline_path,
                                                                 output_path=str(Path(output_file).parent / f"{Path(output_file).stem}_cut_to_grid{Path(output_file).suffix}"))
                input_file_cut = input_file
                input_metadata = raster_metadata(input_file)

                warp_kwargs = {
                    "format": "vrt",
                    "outputType": gdal.gdalconst.GDT_Float32,
                    "warpOptions": wopt,
                    "errorThreshold": 0,
                    "xRes": abs(xres),
                    "yRes": abs(yres),
                    "cutlineDSName": cutline_path,
                    "cropToCutline": True,
                    "coordinateOperation": pipe,
                    "dstNodata": -9999.0,
                }
                if not (crs_utils.multiple_geodetic_crs(self.steps) or crs_utils.multiple_projections(self.steps)):
                    warp_kwargs["outputBounds"] = input_metadata["extent"]

                ds = gdal.Warp(output_vrt, input_file, **warp_kwargs)             

            else:
                # Prepare warp parameters
                warp_kwargs = {
                    "format": "vrt",
                    "outputType": gdal.gdalconst.GDT_Float32,
                    "warpOptions": wopt,
                    "errorThreshold": 0,
                    "xRes": abs(xres),
                    "yRes": abs(yres),
                    "coordinateOperation": pipe,
                    "dstNodata": -9999.0
                }
                if not (crs_utils.multiple_geodetic_crs(self.steps) or crs_utils.multiple_projections(self.steps)):
                    warp_kwargs["outputBounds"] = input_metadata["extent"]
                ds = gdal.Warp(output_vrt, input_file, **warp_kwargs)

            pipe = re.sub(r"\s{2,}", " ", pipe).strip()
            to_wkt = self.crs_to.to_wkt()
            to_wkt = re.sub(r"\s{2,}", " ", to_wkt).strip()
            buffer = io.StringIO()
            sys.stdout = buffer
            pp.show_versions()
            sys.stdout = sys.__stdout__
            pyproj_versions = re.sub(r"\s{2,}", " ", buffer.getvalue()).strip()
            crs_h, crs_v = crs_utils.crs_components(self.crs_to, raise_no_auth=False)

            vyper_meta = {"description": ("This file is the output of a transformation pipeline "
                                          f"executed using Vyperdatum ({version('vyperdatum')})"
                                          " software by NOAA's OCS, NBS branch."),
                          "vyperdatum_version": version("vyperdatum"),
                          "steps": self.steps,
                          "proj_pipeline": pipe,
                          "wkt": to_wkt,
                          "crs_horizontal": crs_h if crs_h else "",
                          "crs_vertical": crs_v if crs_v else "",
                          "pyproj_versions": pyproj_versions,
                          }
            vyper_meta = json.dumps(vyper_meta)
            ds.SetMetadataItem("Vyperdatum_Metadata", vyper_meta)

            # FUSE might have already created a file with the same name, so we need to check
            if os.path.exists(output_file):
                suffix = "_vyperdatum"
                op = Path(output_file)
                new_name = f"{op.stem}{suffix}{op.suffix}"
                output_file = str(op.with_name(new_name))

            cop = ["COMPRESS=DEFLATE"]
            if input_metadata["driver"].lower() == "gtiff":
                cop.append("TILED=YES")
            if input_metadata["driver"].lower() == "bag":
                try:
                    block_size = min(int(input_metadata["block_size"][0][0]),
                                     int(input_metadata["block_size"][0][1]))  # take the smaller block size (x, y)
                    cop.append(f"BLOCK_SIZE={block_size}")
                except Exception as e:
                    logger.warning("Could not parse block size from input raster metadata. "
                                   f"Found invalid block_size value: {input_metadata['block_size'][0]}."
                                   f"\n Exception: {str(e)}")

            output_ds = gdal.Translate(output_file, ds, format=input_metadata["driver"],
                                       outputType=gdal.GDT_Float32,
                                       creationOptions=cop)

            if not os.path.exists(output_file):
                logger.error(f"Output raster was not created: {output_file}")
                logger.error(f"GDAL last error: {gdal.GetLastErrorMsg()}")
                return False

            output_ds = None
            ds = None

            overwrite_with_original(input_file, output_file)
            update_raster_wkt(output_file, to_wkt)
            apply_nbs_band_standards(output_file)
            input_metadata = raster_metadata(input_file)
            output_metadata = raster_metadata(output_file)

            if pre_post_checks:
                raster_utils.raster_post_transformation_checks(source_meta=input_metadata,
                                                               target_meta=output_metadata,
                                                               target_crs=self.crs_to,
                                                               vertical_transform=v_shift
                                                               )
            success = True
            if vdatum_check:
                vdatum_cv, vdatum_df = vdatum_cross_validate(s_wkt=input_metadata["wkt"],
                                                             t_wkt=output_metadata["wkt"],
                                                             n_sample=20,
                                                             s_raster_metadata=input_metadata,
                                                             t_raster_metadata=output_metadata,
                                                             s_point_samples=None,
                                                             t_point_samples=None,
                                                             tolerance=0.3,
                                                             raster_sampling_band=1,
                                                             region=None,
                                                             pivot_h_crs="EPSG:6318",
                                                             s_h_frame=None,
                                                             s_v_frame=None,
                                                             s_h_zone=None,
                                                             t_h_frame=None,
                                                             t_v_frame=None,
                                                             t_h_zone=None
                                                             )
                csv_path = os.path.join(os.path.split(output_file)[0],
                                        os.path.split(output_file)[1].split(".")[0] + "_vdatum_check.csv")
                vdatum_df.to_csv(csv_path, index=False)
                if not vdatum_cv:
                    success = False
                    logger.info(f"{Fore.RED}VDatum API outputs stored at: {csv_path}")
                    print(Style.RESET_ALL)

            # if overview and input_metadata["driver"].lower() == "gtiff":
            #     raster_utils.add_overview(raster_file=output_file,
            #                               compression=input_metadata["compression"]
            #                               )
            #     # raster_utils.add_rat(output_file)

            # Clean up cutline artifacts
            if cutline_path and os.path.exists(cutline_path):
                try:
                    os.remove(cutline_path)
                    os.remove(input_file_cut)
                except Exception as e:
                    logger.warning(f"Could not remove temporary cutline-related artifacts: {e}")
        except Exception as e:
            efile = open(Path(output_file).parent.absolute()/Path(f"{os.path.split(input_file)[1]}_error.txt"), "w")
            efile.write(str(e))
            efile.close()
        finally:
            if os.path.isfile(output_vrt):
                os.remove(output_vrt)
            return success

    def transform_vector(self,
                         input_file: str,
                         output_file: str
                         ) -> bool:
        """
        Transform the gdal-supported input vector file (`input_file`) and store the
        transformed file on the local disk (`output_file`).

        Raises
        -------
        FileNotFoundError:
            If the input vector file is not found.
        NotImplementedError:
            If the input vector file is not supported by gdal.

        Parameters
        -----------
        input_file: str
            Path to the input vector file (gdal supported).
        output_file: str
            Path to the transformed vector file.

        Returns
        --------
        bool:
            True if successful, otherwise False.
        """
        try:
            self._validate_input_file(input_file)
            pathlib.Path(os.path.split(output_file)[0]).mkdir(parents=True, exist_ok=True)
            pbar, success = None, False
            ds = gdal.OpenEx(input_file)
            driver = ogr.GetDriverByName(ds.GetDriver().ShortName)
            inSpatialRef = osr.SpatialReference()
            inSpatialRef.ImportFromWkt(self.crs_from.to_wkt())
            outSpatialRef = osr.SpatialReference()
            outSpatialRef.ImportFromWkt(self.crs_to.to_wkt())
            coordTrans = osr.CoordinateTransformation(inSpatialRef, outSpatialRef)
            inDataSet = driver.Open(input_file)
            if os.path.exists(output_file):
                driver.DeleteDataSource(output_file)
            outDataSet = driver.CreateDataSource(output_file)
            layer_count = inDataSet.GetLayerCount()
            for layer_index in range(layer_count):
                inLayer = inDataSet.GetLayer(layer_index)
                outLayer = outDataSet.CreateLayer(inLayer.GetName(), geom_type=ogr.wkbMultiPolygon)
                inLayerDefn = inLayer.GetLayerDefn()
                for i in range(0, inLayerDefn.GetFieldCount()):
                    fieldDefn = inLayerDefn.GetFieldDefn(i)
                    outLayer.CreateField(fieldDefn)
                outLayerDefn = outLayer.GetLayerDefn()
                inFeature = inLayer.GetNextFeature()
                feature_count = inLayer.GetFeatureCount()
                pbar = tqdm(total=feature_count)
                feature_counter = 0
                while inFeature:
                    geom = inFeature.GetGeometryRef()
                    geom.Transform(coordTrans)
                    outFeature = ogr.Feature(outLayerDefn)
                    outFeature.SetGeometry(geom)
                    for i in range(0, outLayerDefn.GetFieldCount()):
                        outFeature.SetField(outLayerDefn.GetFieldDefn(i).GetNameRef(), inFeature.GetField(i))
                    outLayer.CreateFeature(outFeature)
                    outFeature = None
                    inFeature = inLayer.GetNextFeature()
                    feature_counter += 1
                    pbar.update(1)
                    pbar.set_description(f"Processing Layer {layer_index+1} / {layer_count}")
            inDataSet, outDataSet, ds = None, None, None
            success = True
        finally:
            if pbar:
                pbar.close()
            return success



import os
import glob
import pathlib
from vyperdatum.transformer import Transformer
from vyperdatum.utils.raster_utils import raster_metadata, raster_compress
from osgeo import gdal
import pyproj as pp


def transform(input_file, home_dir):
    options = {"options": ["s_coord_epoch=2010.0", "t_coord_epoch=2010.0"]}
    warp_kwargs_vertical = {
                            "outputType": gdal.gdalconst.GDT_Float32,
                            "srcBands": [1],
                            "dstBands": [1],
                            "warpOptions": ["APPLY_VERTICAL_SHIFT=YES", "SAMPLE_GRID=YES", "SAMPLE_STEPS=ALL"],
                            "errorThreshold": 0,
                            }

    t1 = Transformer(crs_from="EPSG:32619+EPSG:5866",
                     crs_to="EPSG:9755+EPSG:5866",
                     allow_ballpark=False
                     )
    out_file1 = pathlib.Path(input_file).with_stem("_01_" + pathlib.Path(input_file).stem)
    t1.transform_raster(input_file=input_file,
                        output_file=out_file1,
                        apply_vertical=False,
                        warp_kwargs=options
                        )

    t2 = Transformer(crs_from="EPSG:9755",   # must have been EPSG:9755+EPSG:5866, but bug in db/register
                     crs_to="EPSG:6318",     # must have been EPSG:6318+EPSG:5866, but bug in db/register
                     allow_ballpark=False
                     )
    out_file2 = pathlib.Path(input_file).with_stem("_02_" + pathlib.Path(input_file).stem)
    t2.transform_raster(input_file=out_file1,
                        output_file=out_file2,
                        apply_vertical=False,
                        warp_kwargs=options
                        )

    t3 = Transformer(crs_from="EPSG:6318+EPSG:5866",
                     crs_to="EPSG:26919+EPSG:5866",
                     allow_ballpark=False
                     )
    # out_file3 = pathlib.Path(input_file).with_stem("_03_" + pathlib.Path(input_file).stem)
    p = pathlib.Path(input_file)
    xform_dir = os.path.join(home_dir + r"\Manual", p.parent.name)
    os.makedirs(xform_dir, exist_ok=True)
    out_file3 = os.path.join(xform_dir, p.name)
    t3.transform_raster(input_file=out_file2,
                        output_file=out_file3,
                        apply_vertical=False,
                        )

    os.remove(out_file1)
    os.remove(out_file2)
    return out_file3


if __name__ == "__main__":
    # home_dir: path to the dir where the "Original" and "Manual" are stored
    home_dir = r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\PBC\BAG"
    files = glob.glob(home_dir + r"\Original\**\*.bag", recursive=True)
    for i, input_file in enumerate(files):
        print(f"{i+1}/{len(files)}: {input_file}")
        raster_metadata(input_file, verbose=True)
        transformed_file = transform(input_file, home_dir)



from vyperdatum.transformer import Transformer


if __name__ == "__main__":

    input_file = r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\point\MLG\Original\AR_01_BAR_20240117_PR\AR_01_BAR_20240117_PR.XYZ"
    crs_from = "ESRI:103295+NOAA:1502" # MLG depth not usable for now, use NOAA:1502 USACE height
    crs_to = "EPSG:6344+NOAA:101"  # MLLW depth (due to current db issues, I use NOAA:101 and negate to get depth)
    negate_z = True  # MLG depth is negative, NOAA:98 is positive


    tf = Transformer(crs_from=crs_from,
                     crs_to=crs_to,
                    #  steps=steps
                     )
    output_file = input_file.replace("Original", "Manual")
    tf.transform(input_file=input_file,
                 output_file=output_file,
                 pre_post_checks=True,
                 vdatum_check=False,
                 negate_z=negate_z,
                 unit_conversion=0.3048006096
                 )


    # input_file = r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\point\MLG\Original\AR_01_BAR_20240117_PR\AR_01_BAR_20240117_PR.XYZ"
    # crs_from = "EPSG:3452+NOAA:1502" # MLG depth not usable for now, use NOAA:1502 USACE height
    # crs_to = "EPSG:6344+NOAA:100"  # MLLW depth (due to current db issues, I use NOAA:101 and negate to get depth)
    # negate_z = True  # MLG depth is negative, NOAA:98 is positive

    # steps = [{"crs_from": "EPSG:3452", "crs_to": "EPSG:6318", "v_shift": False},
    #          {"crs_from": "EPSG:6318+NOAA:1502", "crs_to": "EPSG:6318+NOAA:101", "v_shift": True},
    #          {"crs_from": "EPSG:6318", "crs_to": "EPSG:6344", "v_shift": False}
    #          ]

    # tf = Transformer(crs_from=crs_from,
    #                  crs_to=crs_to,
    #                  steps=steps
    #                  )
    # output_file = input_file.replace("Original", "Manual")
    # tf.transform(input_file=input_file,
    #              output_file=output_file,
    #              pre_post_checks=True,
    #              vdatum_check=False,
    #              negate_z=negate_z,
    #              unit_conversion=0.3048006096
    #              )



from vyperdatum.transformer import Transformer
from glob import glob
import numpy as np
from osgeo import gdal


gdal.UseExceptions()


crs_from = "EPSG:6348"
crs_to = "EPSG:6348+NOAA:98"

input_files = glob(r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\RSD\ME\Original\**\*.tif", recursive=True)
for i, input_file in enumerate(input_files):
    print(f"Processing ({i}/{len(input_files)}): {input_file}")
    try:
        output_file = input_file.replace("Original", "Manual")
        tf = Transformer(crs_from=crs_from,
                        crs_to=crs_to,
                        )
        tf.transform(input_file=input_file,
                    output_file=output_file,
                    pre_post_checks=True,
                    vdatum_check=True
                    )
    except Exception as e:
        print(f"Error processing {input_file}: {e}")
        continue


from importlib.metadata import version
import logging
import os
import pathlib
import json
import time
import logging.config
from osgeo import gdal
import pyproj as pp
from vyperdatum.db import DB
from vyperdatum.enums import PROJDB, DATUM_DOI
from vyperdatum.utils import assets_util


logger = logging.getLogger("root_logger")


def validate_vyper_grids():
    dir_path = PROJDB.DIR.value
    if dir_path is None:
        raise EnvironmentError("VYPER_GRIDS environment variable is not set. "
                               "Please set it to the directory where the proj database "
                               "and the grid file are located.")

    if not os.path.exists(dir_path):
        raise FileNotFoundError(f"The directory specified by VYPER_GRIDS environment variable does not exist: {dir_path}")
    assert os.environ["VYPER_GRIDS"] == PROJDB.DIR.value
    assert os.environ["PROJ_DATA"] == PROJDB.DIR.value, ("PROJ_DATA environment variable is not set to the correct directory. "
                                                         "It should be set to the same directory specified by VYPER_GRIDS env variable.\n"
                                                         f"Current PROJ_DATA env var: {os.environ['PROJ_DATA']}\n"
                                                         f"Current VYPER_GRIDS env var: {os.environ['VYPER_GRIDS']}")

    selected_grid_files = [
        "us_noaa_nos_underkeel_hydroid-NAD83(2011)_2010.0_(nwldatum_4.7.0_20240621).tif",
        "us_noaa_nos_MLLW-NAD83(2011)_2010.0_(nwldatum_4.7.0_20240621).tif",
        "us_noaa_nos_LWD-NAD83(2011)_2010.0_(nwldatum_4.7.0_20240621).tif",
        "us_noaa_nos_LWD_IGLD85-NAD83(2011)_2010.0_(nwldatum_4.7.0_20240621).tif",
        "us_noaa_nos_survey_hydroid-NAD83(2011)_2010.0_(usace_1.0.0_20250501).tif",
        "us_noaa_nos_CRD-NAD83(2011)_2010.0_(nwldatum_4.7.0_20240621).tif",
        "us_noaa_g2018u0.tif",
        "us_noaa_g2012bu0.tif"]
    missing_grids = []
    for grid in selected_grid_files:
        if not os.path.exists(os.path.join(dir_path, grid)):
            missing_grids.append(grid)
    if len(missing_grids) > 0:
        raise FileNotFoundError(f"The following required grid files are missing in the directory specified by VYPER_GRIDS: {dir_path}\n"
                                f"{', '.join(missing_grids)}\n")
    logger.info(f"Vyperdatum grids directory is set to: {dir_path}")
    return


__version__ = version("vyperdatum")
log_configuration_dict = json.load(
    open(
        pathlib.Path(
            pathlib.Path(__file__).parent, "logging_conf.json"
        )
    )
)
logging.config.dictConfig(log_configuration_dict)
logging.Formatter.converter = time.gmtime

os.environ.update(PROJ_NETWORK="ON")
gdal.UseExceptions()


os.environ["PROJ_DEBUG"] = "2"
os.environ["PROJ_ONLY_BEST_DEFAULT"] = "YES"
os.environ["PROJ_DATA"] = PROJDB.DIR.value  # derived from os.environ.get("VYPER_GRIDS", None)
validate_vyper_grids()
db = DB(db_dir=PROJDB.DIR.value)
assert "NOAA" in pp.database.get_authorities(), ("The authority 'NOAA' not found in proj.db. "
                                                 "Check if the latest database is used.")



import os
import logging
from osgeo import gdal
import laspy
import numpy as np
import pyproj as pp
from vyperdatum.drivers.base import Driver

logger = logging.getLogger("root_logger")
gdal.UseExceptions()


class LAZ(Driver):
    def __init__(self, input_file: str, invalid_error: bool = True) -> None:
        """
        Parameters
        -----------
        input_file: str
            Path to the input laz file.
        invalid_error: bool, default True
            If True, throws an error when the input file has an unexpected format.

        Raises
        -------
        FileNotFoundError:
            If the input file is not found.
        ValueError
            If the passed input file is not recognized as laz file.

        Returns
        -----------
        None
        """
        super().__init__()
        self.input_file = input_file
        if not os.path.isfile(self.input_file):
            raise FileNotFoundError(f"The input file not found at {input_file}.")
        self.is_laz = self.get_points()
        if invalid_error and not self.is_laz:
            msg = (f"The following file is not a valid LAZ file: {self.input_file}")
            logger.exception(msg)
            raise TypeError(msg)
        return

    def get_points(self) -> bool:
        """
        Extract the coordinate points from .LAZ file.

        Returns
        -----------
        bool
            True if the coordinates are found successfully; otherwise False.
        """
        try:
            with laspy.open(self.input_file) as lf:
                points = lf.read()
                self.x = np.array(points.x)
                self.y = np.array(points.y)
                self.z = np.array(points.z)
                valid = True
        except:
            valid = False
        return valid

    def wkt(self) -> str:
        """
        Return the LAZ WKT string.

        Returns
        -----------
        str
            WKT associated with file's CRS.
        """
        with laspy.open(self.input_file) as lf:
            w = lf.header.parse_crs().to_wkt()
        return w

    def transform(self, transformer_instance, vdatum_check: bool) -> bool:
        """
        Apply point transformation on the laz data according to the `transformer_instance`.

        Parameters
        -----------
        transformer_instance: vyperdatum.transformer.Transform
            Instance of the transformer class.

        Returns
        -----------
        bool:
            True if successful, otherwise False.
        """
        lf = laspy.read(self.input_file)
        success, xx, yy, zz = transformer_instance.transform_points(self.x,
                                                                    self.y,
                                                                    self.z,
                                                                    vdatum_check=vdatum_check
                                                                    )
        lf.x, lf.y, lf.z = xx, yy, zz
        lf.header.add_crs(transformer_instance.crs_to)
        lf.write(self.input_file)
        return success

    @property
    def is_valid(self):
        return self.is_laz



import os
import pathlib
import logging
import json
import math
import tempfile
from pathlib import Path
from osgeo import gdal, osr, ogr
import numpy as np
import subprocess
from typing import Union, Optional, Tuple
import pyproj as pp
from vyperdatum.utils.spatial_utils import overlapping_regions, overlapping_extents
from vyperdatum.utils.crs_utils import commandline, pipeline_string, crs_components
from vyperdatum.enums import VDATUM
from vyperdatum.utils import crs_utils

logger = logging.getLogger("root_logger")
gdal.UseExceptions()


def band_stats(band_array: np.ndarray) -> list:
    """
    Return a list containing the min, max, mean, and std of the band array.

    Parameters
    ----------
    band_array: numpy.ndarray
        raster band array

    Returns:
    --------
    list:
        min, max, mean, and std of the band array.
    """
    return [np.nanmin(band_array), np.nanmax(band_array),
            np.nanmean(band_array), np.nanstd(band_array)]


def raster_metadata(raster_file: str, verbose: bool = False) -> dict:
    try:
        metadata = {}
        ds = gdal.Open(raster_file, gdal.GA_ReadOnly)
        srs = ds.GetSpatialRef()
        gdal_metadata = ds.GetMetadata()
        metadata |= {"path": raster_file}
        metadata |= {"description": ds.GetDescription()}
        metadata |= {"driver": ds.GetDriver().ShortName}
        metadata |= {"bands": ds.RasterCount}
        metadata |= {"dimensions": f"{ds.RasterXSize} x {ds.RasterYSize}"}
        metadata |= {"band_no_data": [ds.GetRasterBand(i+1).GetNoDataValue()
                                      for i in range(ds.RasterCount)]}
        metadata |= {"block_size": [ds.GetRasterBand(i+1).GetBlockSize()
                                      for i in range(ds.RasterCount)]}
        metadata |= {"band_descriptions": [ds.GetRasterBand(i+1).GetDescription()
                                           for i in range(ds.RasterCount)]}
        # metadata |= {"band_stats": [band_stats(ds.GetRasterBand(i+1).ReadAsArray())
        #                             for i in range(ds.RasterCount)]}
        metadata |= {"compression": ds.GetMetadata("IMAGE_STRUCTURE").get("COMPRESSION", None)}
        metadata |= {"Vyperdatum_Metadata": json.loads(gdal_metadata.get("Vyperdatum_Metadata", "{}"))}
        metadata |= {"coordinate_epoch": srs.GetCoordinateEpoch()}
        geot = ds.GetGeoTransform()
        res_x, res_y = geot[1], geot[5]
        x_min, y_max = geot[0], geot[3]
        x_max = x_min + res_x * ds.RasterXSize
        y_min = y_max + res_y * ds.RasterYSize
        metadata |= {"geo_transform": geot}
        metadata |= {"extent": [x_min, y_min, x_max, y_max]}
        metadata |= {"resolution": [res_x, res_y]}
        metadata |= {"wkt": srs.ExportToWkt()}
        input_crs = pp.CRS(metadata["wkt"])
        h_ac, v_ac = crs_components(crs=input_crs, raise_no_auth=False)
        metadata |= {"h_authcode": h_ac}
        metadata |= {"v_authcode": v_ac}
        ds = None

        if input_crs.is_compound:
            input_horizontal_crs = pp.CRS(input_crs.sub_crs_list[0])
            input_vertical_crs = pp.CRS(input_crs.sub_crs_list[1])
        else:
            input_horizontal_crs = input_crs
            input_vertical_crs = None

        transformer = pp.Transformer.from_crs(input_horizontal_crs,
                                              "EPSG:6318",
                                              always_xy=True
                                              )
        [[lon_min, lon_max], [lat_min, lat_max]] = transformer.transform([x_min, x_max],
                                                                         [y_min, y_max])
        metadata |= {"geo_extent": [lon_min, lat_min, lon_max, lat_max]}
        metadata |= {"overlapping_regions": overlapping_regions(VDATUM.DIR.value, *metadata["geo_extent"])}
        metadata |= {"overlapping_extents": overlapping_extents(*metadata["geo_extent"])}
        metadata |= {"info": gdal.Info(raster_file, format="json")}

    except Exception as e:
        logger.exception(f"Unable to get raster metadata: {e}")

    if verbose:
        print(f"{'-'*80}\nFile: {pathlib.Path(raster_file).name}"
              f"\n\tInput CRS: {input_crs.name}"
              f"\n\tInput Horizontal Authority: {input_horizontal_crs.to_authority()}"
              f"\n\tInput Vertical Authority: {input_vertical_crs.to_authority() if input_crs.is_compound else None}"
              f"\n\tInput Vertical CRS: {input_vertical_crs if input_crs.is_compound else None}"
              f"\n\tInput Vertical CRS WKT: {input_vertical_crs.to_wkt() if input_crs.is_compound else None}"
              f"\n{'-'*80}\n"
              )
    return metadata


def add_overview(raster_file: str, compression: str = "", embedded: bool = True) -> None:
    """
    Add overview bands to a raster file with no existing overviews.

    parameters
    ----------
    raster_file: str
        Absolute full path to the raster file.
    compression: str
        The name of compression algorithm.
    embedded: bool, default=True
        If True, the overviews will be embedded in the file, otherwise stored externally.
    """
    try:
        ds = gdal.Open(raster_file, gdal.GA_Update if embedded else gdal.GA_ReadOnly)
        if compression:
            gdal.SetConfigOption("COMPRESS_OVERVIEW", compression)
    finally:
        ds.BuildOverviews("NEAREST", [2, 4, 16], gdal.TermProgress_nocb)
        ds = None
    return


def add_rat(raster: str) -> None:
    """
    Add Raster Attribute Table (RAT) to all bands of a raster file.

    parameters
    ----------
    raster_file: str
        Absolute full path to the raster file.
    """
    ds = gdal.Open(raster)
    for i in range(ds.RasterCount):
        rat = gdal.RasterAttributeTable()
        rat.CreateColumn("VALUE", gdal.GFT_Real, gdal.GFU_Generic)
        rat.CreateColumn("COUNT", gdal.GFT_Integer, gdal.GFU_Generic)
        band = ds.GetRasterBand(i+1)
        unique, counts = np.unique(band.ReadAsArray(), return_counts=True)
        for i in range(len(unique)):
            rat.SetValueAsDouble(i, 0, float(unique[i]))
            rat.SetValueAsInt(i, 1, int(counts[i]))
        band.SetDefaultRAT(rat)
    ds = None
    return


def set_nodatavalue(raster_file: str,
                    band_nodatavalue: Union[list[tuple[int, float]], float],
                    ) -> None:
    """
    Change the NoDataValue of the raster file bands.

    parameters
    ----------
    raster_file: str
        Absolute full path to the raster file.
    band_nodatavalue: Union[list[tuple[int, float]], float]
        A list of tuples: (band_index, NoDataValue).
        If a single float is passed, all bands will be affected.
    """
    if isinstance(band_nodatavalue, float):
        ds = gdal.Open(raster_file)
        band_nodatavalue = [(b, band_nodatavalue) for b in range(1, ds.RasterCount+1)]
        ds = None
    ds = gdal.Open(raster_file, gdal.GA_Update)
    for b, nodv in band_nodatavalue:
        band = ds.GetRasterBand(b)
        bar = band.ReadAsArray()
        bar[np.where(bar == band.GetNoDataValue())] = nodv
        band.WriteArray(bar)
        ds.GetRasterBand(b).SetNoDataValue(nodv)
        band = None
    ds = None
    return


def unchanged_to_nodata(src_raster_file: str,
                        xform_raster_file: str,
                        xform_band: int,
                        ) -> None:
    """
    Compare the `xform_band` values of the `src_raster_file` and `xform_raster_file`.
    Change the values to NoDataValue if the transformed value is the same as the
    original value (indicating that the transformation has failed).
    Currentl, PROJ keep the source raster unchanged when fails to apply the transformation.
    The transformation fails when the source data is outside any of the underlying
    transformation grids. This function is meant to replace the failed transformation
    points with NoDataValue.  

    parameters
    ----------
    src_raster_file: str
        Absolute full path to the source raster file.
    xform_raster_file: str
        Absolute full path to the transformed raster file.
    xform_band: int
        The reference band index that is used for comparison between the source and transformed file.
    """
    src_ds = gdal.Open(src_raster_file)
    xform_ds = gdal.Open(xform_raster_file, gdal.GA_Update)
    src_band = src_ds.GetRasterBand(xform_band)
    xform_band = xform_ds.GetRasterBand(xform_band)
    ndv = src_band.GetNoDataValue()
    sar, xar = src_band.ReadAsArray(), xform_band.ReadAsArray()
    unchanged_mask = np.where(np.abs(sar - xar) < 0.01)
    src_band, xform_band = None, None
    for b in range(1, xform_ds.RasterCount+1):
        band = xform_ds.GetRasterBand(b)
        bar = band.ReadAsArray()
        bar[unchanged_mask] = ndv
        band.WriteArray(bar)
        xform_ds.GetRasterBand(b).SetNoDataValue(ndv)
        band = None
    src_ds, xform_ds = None, None
    return


def raster_compress(raster_file_path: str,
                    output_file_path: str,
                    format: str,
                    compression: str
                    ):
    """
    Compress raster file.

    Parameters
    ----------
    raster_file_path: str
        absolute path to the input raster file.
    output_file_path: str
        absolute path to the compressed output raster file.
    format: str
        raster file format.
    compression: str
        compression algorithm.
    """
    translate_kwargs = {"format": format,
                        "creationOptions": [f"COMPRESS={compression}"]
                        }
    gtiff_co = [f"BIGTIFF={'YES' if os.path.getsize(raster_file_path) > 3e9 else 'IF_NEEDED'}",
                "TILED=YES"]

    if format.lower() == "gtiff":
        translate_kwargs["creationOptions"].extend(gtiff_co)
    gdal.Translate(output_file_path, raster_file_path, **translate_kwargs)
    return


def preserve_raster_size(input_file: str,
                         output_file: str
                         ):
    """
    Resize the `output_file` raster dimensions to those of the input `input_file`.

    Parameters
    ----------
    input_file: str
        absolute path to the input raster file.
    output_file: str
        absolute path to the compressed output raster file.
    """
    ds_in = gdal.Open(input_file)
    ds_out = gdal.Open(output_file)
    w_in, h_in = ds_in.RasterXSize, ds_in.RasterYSize
    w_out, h_out = ds_out.RasterXSize, ds_out.RasterYSize
    ds_in, ds_out = None, None
    if w_in != w_out or h_in != h_out:
        output_file_copy = str(output_file)+".tmp"
        os.rename(output_file, output_file_copy)
        gdal.Translate(output_file, output_file_copy, width=w_in, height=h_in)
        os.remove(output_file_copy)
    return


def crs_to_code_auth(crs: pp.CRS) -> Optional[str]:
    """
    Return CRS string representation in form of code:authority

    Raises
    -------
    ValueError:
        If either code or authority of the crs (or its sub_crs) can not be determined.

    Returns
    --------
    str:
        crs string in form of code:authority
    """
    def get_code_auth(_crs: pp.CRS):
        if _crs.to_authority(min_confidence=100):
            return ":".join(_crs.to_authority(min_confidence=100))
        raise ValueError(f"Unable to produce authority name and code for this crs:\n{_crs}")

    if crs.is_compound:
        hcrs = pp.CRS(crs.sub_crs_list[0])
        vcrs = pp.CRS(crs.sub_crs_list[1])
        code_auth = f"{get_code_auth(hcrs)}+{get_code_auth(vcrs)}"
    else:
        code_auth = get_code_auth(crs)
    return code_auth


def push_pop_at_vshift(pipe: str) -> str:
    """
    Sandwich the vgrid_shift operation with push/pop of the horizontal CRS components.

    Parameters
    ----------
    pipe: proj pipeline string.
    """
    horz_push = " +step +proj=push +v_1 +v_2 "
    horz_pop = " +step +proj=pop +v_1 +v_2 "

    vgird_delim = "+step +inv +proj=vgridshift"
    vgrid_splits = pipe.split(vgird_delim)
    if len(vgrid_splits) != 2:
        raise ValueError(f"Pipeline split by '{vgird_delim}' must result in "
                         f"exactly 2 parts but got {len(vgrid_splits)} parts:\n {vgrid_splits}")
    hydroid_delim = ".tif"
    hydroid_splits = vgrid_splits[-1].split(hydroid_delim)

    sand_pipe = vgrid_splits[0] + horz_push + vgird_delim + " " + (hydroid_delim + " ").join(hydroid_splits[:-1]) + hydroid_delim + " "

    step_delim = "+step"
    step_splits = hydroid_splits[-1].split(step_delim)

    sand_pipe += step_splits[0] + horz_pop + step_delim + step_delim.join(step_splits[1:])
    return sand_pipe


def raster_pre_transformation_checks(source_meta: dict, source_crs: Union[pp.CRS, str]):
    """
    Run a number of sanity checks on the source raster file, before transformation.
    Warns if a check fails.

    Parameters
    ----------
    source_meta: dict
        Source raster metadata generated by `raster_metadata` function.

    Returns
    ----------
    bool
        Returns True if all checks pass, otherwise False.
    """
    passed = True
    if ~isinstance(source_crs, pp.CRS):
        source_crs = pp.CRS(source_crs)
    source_auth = source_crs.to_authority()
    raster_auth = pp.CRS(source_meta["wkt"]).to_authority()
    if source_auth != raster_auth:
        passed = False
        logger.warning("The expected authority code/name of the "
                       f"input raster file is {raster_auth}, but received {source_auth}"
                       )
    if source_meta["bands"] != 3:
        passed = False
        logger.warning("Number of bands in the raster file: "
                       f"{source_meta['bands']}. NBS rasters typically contain 3 bands")
    if len(source_meta["overlapping_regions"]) != 1:
        passed = False
        logger.warning("The raster is not overlapping with a single region. "
                       f"The overlapping regions: ({source_meta['overlapping_regions']}).")
    return passed


def raster_post_transformation_checks(source_meta: dict,
                                      target_meta: dict,
                                      target_crs: Union[pp.CRS, str],
                                      vertical_transform: bool
                                      ):
    """
    Run a number of sanity checks on the transformed raster file.
    Warns if a check fails.

    Parameters
    ----------
    source_meta: dict
        Source raster metadata generated by `raster_metadata` function.
    target_meta: dict
        Target raster metadata generated by `raster_metadata` function.
    target_crs: pyproj.crs.CRS or input used to create one
        The expected CRS object for the target raster file.
    vertical_transform: bool
        True if it's a vertical transformation, otherwise False.

    Returns
    ----------
    bool
        Returns True if all checks pass, otherwise False.
    """
    if ~isinstance(target_crs, pp.CRS):
        target_crs = pp.CRS(target_crs)
    passed = True
    target_auth = target_crs.to_authority()
    transformed_auth = pp.CRS(target_meta["wkt"]).to_authority()
    if target_auth != transformed_auth:
        passed = False
        logger.warning("The expected authority code/name of the "
                       f"transformed raster is {target_auth}, but received {transformed_auth}"
                       )
    if source_meta["bands"] != target_meta["bands"]:
        passed = False
        logger.warning("Number of bands in the source file "
                       f"({source_meta['bands']}) doesn't match target ({target_meta['bands']}).")

    if source_meta["driver"] != target_meta["driver"]:
        passed = False
        logger.warning("The driver of the source file "
                       f"({source_meta['driver']}) doesn't match target ({target_meta['driver']}).")

    if vertical_transform:
        if source_meta["dimensions"] != target_meta["dimensions"]:
            passed = False
            logger.warning("The source file band dimensions "
                           f" ({source_meta['dimensions']}) don't match those of the "
                           f"transformed file ({target_meta['dimensions']}).")
        if source_meta["resolution"][0] != target_meta["resolution"][0] or source_meta["resolution"][1] != target_meta["resolution"][1]:
            passed = False
            logger.warning("The source file pixel size "
                           f" ({source_meta['resolution']}) don't match those of the "
                           f"transformed file ({target_meta['resolution']}).")
    return passed


def update_raster_wkt(input_file: str, wkt: str) -> None:
    """
    Assign a new WKT to a GeoTIFF in-place, or to a BAG by recreating the file.
    WARNING: This only *assigns* the SRS. If you're actually changing CRS
    (e.g., degrees->meters), you must reproject with gdal.Warp instead.

    Parameters
    -----------
    input_file: str
        Absolute path to the input raster file.
    wkt: str
        New WKT to update the raster file.    
    """
    if not os.path.exists(input_file):
        err_msg = f"Trying to update WKT, but the input raster file {input_file} does not exist."
        logger.error(err_msg)
        raise FileNotFoundError(err_msg)

    ds = gdal.Open(input_file, gdal.GA_ReadOnly)
    if ds is None:
        raise RuntimeError(f"GDAL cannot open: {input_file}")

    drv = ds.GetDriver().ShortName.upper()

    # GeoTIFF: update in place
    if drv == "GTIFF":
        ds = None
        upd = gdal.Open(input_file, gdal.GA_Update)
        if upd is None:
            raise RuntimeError("Failed to reopen GeoTIFF in update mode.")
        upd.SetProjection(wkt)
        upd.FlushCache()
        upd = None
        return

    # BAG: recreate with -a_srs
    if drv == "BAG":
        tmp_out = input_file + ".tmp.bag"

        ds_bs = gdal.Open(input_file)
        bs = min(ds_bs.GetRasterBand(1).GetBlockSize())
        ds_bs = None

        # Build gdal.Translate options equivalent to: -of BAG -a_srs "<wkt>"
        topts = gdal.TranslateOptions(
            options=["-of", "BAG", "-a_srs", wkt, "-co", f"BLOCK_SIZE={bs}"]
        )
        out = gdal.Translate(tmp_out, ds, options=topts)
        if out is None:
            raise RuntimeError("gdal.Translate failed when recreating BAG.")
        out = None
        ds = None

        os.replace(tmp_out, input_file)
        return

    # Other drivers: try in-place, else fall back to copy-on-write via Translate
    ds = None
    upd = gdal.Open(input_file, gdal.GA_Update)
    if upd:
        upd.SetProjection(wkt)
        upd.FlushCache()
        upd = None
    else:
        tmp_out = input_file + ".tmp"
        topts = gdal.TranslateOptions(options=["-a_srs", wkt])
        out = gdal.Translate(tmp_out, input_file, options=topts)
        if out is None:
            raise RuntimeError(f"Driver {drv} does not support updating SRS.")
        out = None
        os.replace(tmp_out, input_file)
    return


def overwrite_with_original(input_file: str, output_file: str) -> None:
    """
    Overwrite the non-elevation bands in the output file with the original input
    file band arrays. If an uncertainty band exists, it will be masked so that
    uncertainty is only present where elevation is valid (not NoData).
    """
    ds_in = gdal.Open(input_file, gdal.GA_ReadOnly)
    if ds_in is None:
        raise ValueError(f"Failed to open input_file: {input_file}")

    if ds_in.RasterCount < 2:
        ds_in = None
        return

    elevation_band, uncertainty_band = None, None
    for i in range(1, ds_in.RasterCount + 1):
        band = ds_in.GetRasterBand(i)
        desc = (band.GetDescription() or "").strip().lower()
        if desc in ["elevation", "dem"]:
            elevation_band = i
        if desc in ["uncertainty", "tvu"]:
            uncertainty_band = i

    if elevation_band is None:
        ds_in = None
        raise ValueError(
            "No elevation band found in the input raster file. "
            "Please provide the index of the elevation band."
        )

    ds_out = gdal.Open(output_file, gdal.GA_ReadOnly)
    if ds_out is None:
        ds_in = None
        raise ValueError(f"Failed to open output_file: {output_file}")

    input_metadata = raster_metadata(input_file)
    driver = gdal.GetDriverByName(input_metadata["driver"])

    mem_path = f"/vsimem/{os.path.splitext(os.path.basename(output_file))[0]}.tiff"
    ds_temp = driver.Create(
        mem_path,
        ds_in.RasterXSize,
        ds_in.RasterYSize,
        ds_in.RasterCount,
        gdal.GDT_Float32
    )
    ds_temp.SetGeoTransform(ds_out.GetGeoTransform())
    ds_temp.SetProjection(ds_out.GetProjection())

    # --- Read elevation (from ds_out) once and build a validity mask ---
    elev_out_band = ds_out.GetRasterBand(elevation_band)
    elev_arr = elev_out_band.ReadAsArray()
    elev_nodata = elev_out_band.GetNoDataValue()

    if elev_nodata is None:
        # Fall back: treat non-finite as invalid
        valid_elev = np.isfinite(elev_arr)
    else:
        if np.isnan(elev_nodata):
            valid_elev = ~np.isnan(elev_arr)
        else:
            valid_elev = (elev_arr != elev_nodata) & np.isfinite(elev_arr)

    # --- Write all bands; mask uncertainty where elevation is invalid ---
    for b in range(1, ds_in.RasterCount + 1):
        out_band = ds_temp.GetRasterBand(b)

        if b == elevation_band:
            # Write transformed elevation from ds_out
            out_band.SetDescription("elevation")
            if elev_nodata is not None:
                out_band.SetNoDataValue(float(elev_nodata))
            out_band.WriteArray(elev_arr.astype(np.float32, copy=False))

        elif uncertainty_band is not None and b == uncertainty_band:
            in_unc_band = ds_in.GetRasterBand(b)
            unc_arr = in_unc_band.ReadAsArray().astype(np.float32, copy=False)

            # Choose an output NoData for uncertainty
            unc_nodata_in = in_unc_band.GetNoDataValue()
            if unc_nodata_in is not None:
                unc_nodata_out = float(unc_nodata_in)
            elif elev_nodata is not None and np.isfinite(elev_nodata):
                unc_nodata_out = float(elev_nodata)
            else:
                unc_nodata_out = -9999.0

            # Mask uncertainty wherever elevation is NoData
            unc_masked = np.array(unc_arr, copy=True)
            unc_masked[~valid_elev] = unc_nodata_out

            out_band.SetDescription("uncertainty")
            out_band.SetNoDataValue(unc_nodata_out)
            out_band.WriteArray(unc_masked)

        else:
            # Copy any other non-elevation bands from the input (mask at no elevation points)
            in_band = ds_in.GetRasterBand(b)
            arr = in_band.ReadAsArray()
            if elev_nodata is not None and np.isfinite(elev_nodata):
                unc_nodata_out = float(elev_nodata)
            else:
                unc_nodata_out = -9999.0
            arr_masked = np.array(arr, copy=True)
            arr_masked[~valid_elev] = unc_nodata_out
            out_band.SetDescription(in_band.GetDescription())
            nd = in_band.GetNoDataValue()
            if nd is not None:
                out_band.SetNoDataValue(float(nd))
            out_band.WriteArray(arr_masked.astype(np.float32, copy=False))

    ds_in, ds_out = None, None
    ds_temp.FlushCache()

    cop = ["COMPRESS=DEFLATE"]
    if input_metadata["driver"].lower() == "gtiff":
        cop.extend(["TILED=YES"])

    gdal.Translate(
        output_file,
        ds_temp,
        format=input_metadata["driver"],
        outputType=gdal.GDT_Float32,
        creationOptions=cop
    )

    ds_temp = None
    gdal.Unlink(mem_path)


def update_stats(input_file):
    """
    Update statistics for the raster file.

    Parameters:
        input_file (str): Path to the raster file.
    """
    ds = gdal.Open(input_file, gdal.GA_Update)
    for i in range(1, ds.RasterCount + 1):
        ds.GetRasterBand(i).ComputeStatistics(False)
    ds = None
    return


def apply_nbs_bandnames(input_file):
    """
    Apply NBS band names to the raster file.
    Sets the first band description to 'elevation' and the second band
    description to 'uncertainty' if the raster has more than one band.

    Parameters:
        input_file (str): Path to the raster file.
    """
    ds = gdal.Open(input_file, gdal.GA_Update)
    if ds is None:
        err_msg = f"Could not open file: {input_file}"
        logger.error(err_msg)
        raise RuntimeError(err_msg)

    band = ds.GetRasterBand(1)
    band.SetDescription("elevation")

    if ds.RasterCount > 1:
        band = ds.GetRasterBand(2)
        band.SetDescription("uncertainty")

    band.FlushCache()
    ds = None
    return


def add_uncertainty_band(input_file: str) -> None:
    src_ds = gdal.Open(input_file)
    if src_ds is None:
        raise RuntimeError(f"Could not open {input_file}")

    # assumes uncertainty band already exists
    if src_ds.RasterCount > 1:
        return

    input_metadata = raster_metadata(input_file)
    elev_band = src_ds.GetRasterBand(1)
    nodata_val = elev_band.GetNoDataValue()
    elevation = elev_band.ReadAsArray().astype(np.float32)

    if nodata_val is not None:
        mask = (elevation == nodata_val)
    else:
        mask = np.zeros_like(elevation, dtype=bool)

    uncertainty = 1 + 0.02 * elevation
    uncertainty[mask] = nodata_val if nodata_val is not None else 0

    cols = src_ds.RasterXSize
    rows = src_ds.RasterYSize
    count = src_ds.RasterCount
    geotransform = src_ds.GetGeoTransform()
    projection = src_ds.GetProjection()
    dtype = gdal.GDT_Float32

    # Temp uncompressed file with extra band
    tmp_uncompressed = tempfile.NamedTemporaryFile(suffix=".tif", delete=False)
    tmp_uncompressed_path = tmp_uncompressed.name
    tmp_uncompressed.close()

    driver = gdal.GetDriverByName("GTiff")
    raw_ds = driver.Create(tmp_uncompressed_path, cols, rows, count + 1, dtype)
    raw_ds.SetGeoTransform(geotransform)
    raw_ds.SetProjection(projection)

    for i in range(count):
        band = src_ds.GetRasterBand(i + 1)
        data = band.ReadAsArray()
        out_band = raw_ds.GetRasterBand(i + 1)
        out_band.WriteArray(data)
        out_band.SetDescription(band.GetDescription())
        nd = band.GetNoDataValue()
        if nd is not None:
            out_band.SetNoDataValue(nd)

    # Add uncertainty band
    unc_band = raw_ds.GetRasterBand(count + 1)
    unc_band.WriteArray(uncertainty)
    unc_band.SetDescription("uncertainty")
    if nodata_val is not None:
        unc_band.SetNoDataValue(nodata_val)
    unc_band.FlushCache()

    raw_ds = None
    src_ds = None

    cop = ["COMPRESS=DEFLATE"]
    if input_metadata["driver"].lower() == "gtiff":
        cop.extend(["TILED=YES", "BIGTIFF=IF_SAFER"])
    gdal.Translate(
        destName=input_file,
        srcDS=tmp_uncompressed_path,
        creationOptions=cop,
        format=input_metadata["driver"]
    )

    os.remove(tmp_uncompressed_path)
    return


def apply_nbs_band_standards(input_file: str) -> None:
    """
    Apply NBS band standards to the raster file.
    Sets the first band description to 'elevation' and the second band
    description to 'uncertainty' if the raster has more than one band.
    Adds uncertainty band if it does not exist, and updates statistics.

    Parameters:
        input_file (str): Path to the raster file.
    """
    if not os.path.exists(input_file):
        err_msg = f"Raster file {input_file} does not exist."
        logger.error(err_msg)
        raise FileNotFoundError(err_msg)
    ds = gdal.Open(input_file)
    if ds.GetDriver().ShortName.lower() != "gtiff":
        return
    
    # apply_nbs_bandnames(input_file)
    add_uncertainty_band(input_file)
    update_stats(input_file)
    return


def create_cutline_from_grid(grid_path: str,
                             output_cutline: str,
                             input_extent=None,
                             input_wkt: str = None):
    """
    Create a cutline polygon from a GTG grid's valid (non-nodata) area,
    restricted to the vicinity of `input_extent` (bbox) when provided.
    """
    _logger = globals().get("logger", None)

    def _log(level: str, msg: str):
        if _logger is not None:
            getattr(_logger, level)(msg)
        else:
            print(f"{level.upper()}: {msg}")

    def _inv_geotransform(gt):
        """
        GDAL Python bindings differ by version:
          - Some return (ok, inv_gt)
          - Some return inv_gt (6-tuple) directly
        Normalize to inv_gt (6-tuple) and raise on failure.
        """
        inv = gdal.InvGeoTransform(gt)

        # (ok, inv_gt)
        if isinstance(inv, tuple) and len(inv) == 2 and isinstance(inv[0], (bool, int)):
            ok, inv_gt = inv
            if not ok:
                raise RuntimeError("gdal.InvGeoTransform failed")
            return tuple(inv_gt)

        # direct 6-tuple
        if isinstance(inv, (tuple, list)) and len(inv) == 6:
            return tuple(inv)

        raise RuntimeError(f"Unexpected gdal.InvGeoTransform return: {type(inv)} {inv}")

    def _ds_extent(ds):
        gt = ds.GetGeoTransform()
        x0, px_w, _, y0, _, px_h = gt
        w, h = ds.RasterXSize, ds.RasterYSize
        xmin = x0
        xmax = x0 + px_w * w
        ymax = y0
        ymin = y0 + px_h * h
        if xmin > xmax:
            xmin, xmax = xmax, xmin
        if ymin > ymax:
            ymin, ymax = ymax, ymin
        return (xmin, ymin, xmax, ymax)

    def _bbox_intersection(a, b):
        ax0, ay0, ax1, ay1 = a
        bx0, by0, bx1, by1 = b
        ix0 = max(ax0, bx0)
        iy0 = max(ay0, by0)
        ix1 = min(ax1, bx1)
        iy1 = min(ay1, by1)
        if ix0 >= ix1 or iy0 >= iy1:
            return None
        return (ix0, iy0, ix1, iy1)

    def _bbox_to_srcwin(ds, bbox, pad_pixels=3):
        ds_ext = _ds_extent(ds)
        bb = _bbox_intersection(bbox, ds_ext)
        if bb is None:
            return None
        xmin, ymin, xmax, ymax = bb

        inv_gt = _inv_geotransform(ds.GetGeoTransform())

        # top-left and bottom-right in map coords
        px0, py0 = gdal.ApplyGeoTransform(inv_gt, xmin, ymax)
        px1, py1 = gdal.ApplyGeoTransform(inv_gt, xmax, ymin)

        col0 = int(math.floor(min(px0, px1))) - pad_pixels
        col1 = int(math.ceil(max(px0, px1))) + pad_pixels
        row0 = int(math.floor(min(py0, py1))) - pad_pixels
        row1 = int(math.ceil(max(py0, py1))) + pad_pixels

        col0 = max(0, min(col0, ds.RasterXSize))
        col1 = max(0, min(col1, ds.RasterXSize))
        row0 = max(0, min(row0, ds.RasterYSize))
        row1 = max(0, min(row1, ds.RasterYSize))

        xsize = col1 - col0
        ysize = row1 - row0
        if xsize <= 0 or ysize <= 0:
            return None
        return (col0, row0, xsize, ysize)

    def _srs_from_wkt(wkt: str):
        srs = osr.SpatialReference()
        srs.ImportFromWkt(wkt)
        try:
            srs.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)
        except Exception:
            pass
        return srs

    def _reproject_bbox(bbox, src_srs, dst_srs):
        xmin, ymin, xmax, ymax = bbox
        ct = osr.CoordinateTransformation(src_srs, dst_srs)
        pts = [
            ct.TransformPoint(xmin, ymin),
            ct.TransformPoint(xmin, ymax),
            ct.TransformPoint(xmax, ymin),
            ct.TransformPoint(xmax, ymax),
        ]
        xs = [p[0] for p in pts]
        ys = [p[1] for p in pts]
        return (min(xs), min(ys), max(xs), max(ys))

    def _sample_validity(ds, bbox, nodata):
        xmin, ymin, xmax, ymax = bbox
        xs = [xmin, (xmin + xmax) / 2.0, xmax]
        ys = [ymin, (ymin + ymax) / 2.0, ymax]
        band = ds.GetRasterBand(1)
        inv_gt = _inv_geotransform(ds.GetGeoTransform())

        valid = 0
        total = 0
        for x in xs:
            for y in ys:
                total += 1
                px, py = gdal.ApplyGeoTransform(inv_gt, x, y)
                col = int(round(px))
                row = int(round(py))
                if col < 0 or row < 0 or col >= ds.RasterXSize or row >= ds.RasterYSize:
                    continue
                arr = band.ReadAsArray(col, row, 1, 1)
                if arr is None:
                    continue
                v = float(arr[0, 0])
                if np.isnan(v):
                    continue
                if nodata is not None and np.isclose(v, nodata):
                    continue
                valid += 1
        return valid, total

    try:
        gdal.UseExceptions()
        ogr.UseExceptions()

        container = gdal.Open(grid_path)
        if container is None:
            _log("error", f"Could not open grid: {grid_path}")
            return None

        candidates = []
        subdatasets = container.GetSubDatasets() or []
        if subdatasets:
            _log("info", f"GTG file detected with {len(subdatasets)} subgrids")
            for idx, (name, desc) in enumerate(subdatasets):
                ds = gdal.Open(name)
                if ds is None:
                    continue
                ext = _ds_extent(ds)
                gt = ds.GetGeoTransform()
                px_area = abs(gt[1] * gt[5]) if gt else float("inf")
                candidates.append((idx, name, desc, ds, ext, px_area))
        else:
            ds = container
            ext = _ds_extent(ds)
            gt = ds.GetGeoTransform()
            px_area = abs(gt[1] * gt[5]) if gt else float("inf")
            candidates.append((0, grid_path, "", ds, ext, px_area))

        if not candidates:
            _log("error", "No readable subdatasets found in grid.")
            return None

        grid_wkt = candidates[0][3].GetProjection()
        grid_srs = _srs_from_wkt(grid_wkt) if grid_wkt else None

        bbox_grid = None
        if input_extent is not None:
            bbox_in = tuple(float(x) for x in input_extent)
            if input_wkt and grid_srs:
                in_srs = _srs_from_wkt(input_wkt)
                bbox_grid = _reproject_bbox(bbox_in, in_srs, grid_srs)
            else:
                bbox_grid = bbox_in

        chosen = candidates[0]
        if bbox_grid is not None:
            scored = []
            for idx, name, desc, ds, ext, px_area in candidates:
                inter = _bbox_intersection(bbox_grid, ext)
                if inter is None:
                    continue
                band = ds.GetRasterBand(1)
                nodata = band.GetNoDataValue()
                vcount, _ = _sample_validity(ds, inter, nodata)
                area = abs((ext[2] - ext[0]) * (ext[3] - ext[1]))
                scored.append((vcount, -area, -1.0 / (px_area + 1e-30), idx))
            if scored:
                scored.sort(reverse=True)
                best_idx = scored[0][3]
                chosen = next(c for c in candidates if c[0] == best_idx)
                _log("info", f"Selected subgrid {chosen[0]} using valid-sample scoring. extent={chosen[4]}")
            else:
                # fallback: max overlap
                best = None
                for cand in candidates:
                    idx, name, desc, ds, ext, px_area = cand
                    inter = _bbox_intersection(bbox_grid, ext)
                    if inter is None:
                        continue
                    overlap = (inter[2] - inter[0]) * (inter[3] - inter[1])
                    if best is None or overlap > best[0]:
                        best = (overlap, cand)
                if best:
                    chosen = best[1]
                    _log("warning", f"Fell back to max-overlap subgrid {chosen[0]}")

        idx, chosen_name, chosen_desc, grid_ds, grid_ext, _ = chosen

        if bbox_grid is not None:
            inter = _bbox_intersection(bbox_grid, grid_ext)
            if inter is None:
                _log("error", f"Input bbox does not overlap chosen subgrid extent. bbox={bbox_grid} subgrid={grid_ext}")
                return None
            target_bbox = inter
        else:
            target_bbox = grid_ext
            _log("warning", "No input_extent provided; polygonizing the full grid can be slow/huge.")

        srcwin = _bbox_to_srcwin(grid_ds, target_bbox, pad_pixels=5)
        if srcwin is None:
            _log("error", f"Crop window collapsed. bbox={target_bbox} -> srcwin=None. grid_ext={grid_ext}")
            return None

        xoff, yoff, xsize, ysize = srcwin
        _log("info", f"Cropping grid for polygonize: srcWin={srcwin} (bbox={target_bbox})")

        cropped = gdal.Translate("", grid_ds, format="MEM", srcWin=[xoff, yoff, xsize, ysize])
        if cropped is None:
            _log("error", "gdal.Translate failed while cropping grid.")
            return None

        band = cropped.GetRasterBand(1)
        nodata = band.GetNoDataValue()
        arr = band.ReadAsArray()
        if arr is None:
            _log("error", "Failed to read cropped grid band.")
            return None
        if nodata is None:
            nodata = -32768.0

        valid_mask = (~np.isnan(arr)) & (~np.isclose(arr.astype("float64"), float(nodata)))
        valid_count = int(valid_mask.sum())
        total_count = int(valid_mask.size)
        _log("info", f"Valid cells (cropped): {valid_count} / {total_count} ({valid_count/total_count*100.0:.2f}%)")
        if valid_count == 0:
            _log("error", "No valid grid cells under input extent (coverage or axis-order issue).")
            return None

        # Build a Byte mask dataset for polygonize
        mem = gdal.GetDriverByName("MEM")
        mask_ds = mem.Create("", xsize, ysize, 1, gdal.GDT_Byte)
        mask_ds.SetGeoTransform(cropped.GetGeoTransform())
        if grid_wkt:
            mask_ds.SetProjection(grid_wkt)
        mask_band = mask_ds.GetRasterBand(1)
        mask_band.WriteArray(valid_mask.astype("uint8"))
        mask_band.FlushCache()

        # Polygonize into an in-memory vector layer
        vmem_drv = ogr.GetDriverByName("Memory")
        vmem_ds = vmem_drv.CreateDataSource("mem_cutline")
        srs = _srs_from_wkt(grid_wkt) if grid_wkt else None
        vmem_lyr = vmem_ds.CreateLayer("cutline", srs=srs, geom_type=ogr.wkbPolygon)
        vmem_lyr.CreateField(ogr.FieldDefn("DN", ogr.OFTInteger))

        _log("info", "Polygonizing mask (in-memory)...")
        gdal.Polygonize(mask_band, mask_band, vmem_lyr, 0, ["8CONNECTED=8"])
        feat_count = vmem_lyr.GetFeatureCount()
        _log("info", f"Polygonize created {feat_count} features")

        vmem_lyr.SetAttributeFilter("DN = 1")
        union_geom = None
        for f in vmem_lyr:
            g = f.GetGeometryRef()
            if g is None:
                continue
            g2 = g.Clone()
            union_geom = g2 if union_geom is None else union_geom.Union(g2)

        if union_geom is None:
            _log("error", "Failed to build union geometry from polygonized features.")
            return None

        # Conservative simplify
        if srs and srs.IsGeographic():
            tol = 0.0005
        else:
            tol = 50.0
        try:
            union_geom = union_geom.Simplify(tol)
        except Exception:
            pass

        # Output driver
        ext = os.path.splitext(output_cutline)[1].lower()
        if ext == ".gpkg":
            drv_name = "GPKG"
        elif ext == ".shp":
            drv_name = "ESRI Shapefile"
        elif ext in (".json", ".geojson"):
            drv_name = "GeoJSON"
        else:
            output_cutline = os.path.splitext(output_cutline)[0] + ".gpkg"
            drv_name = "GPKG"
            _log("warning", f"Unrecognized cutline extension; writing GeoPackage instead: {output_cutline}")

        out_drv = ogr.GetDriverByName(drv_name)
        if out_drv is None:
            _log("error", f"OGR driver not available: {drv_name}")
            return None
        if os.path.exists(output_cutline):
            out_drv.DeleteDataSource(output_cutline)

        # For GeoJSON: ask GDAL to write CRS to avoid ambiguous interpretation in some pipelines
        if drv_name == "GeoJSON":
            gdal.SetConfigOption("GDAL_GEOJSON_WRITE_CRS", "YES")

        out_ds = out_drv.CreateDataSource(output_cutline)
        out_lyr = out_ds.CreateLayer("cutline", srs=srs, geom_type=ogr.wkbPolygon)
        out_lyr.CreateField(ogr.FieldDefn("DN", ogr.OFTInteger))

        feat = ogr.Feature(out_lyr.GetLayerDefn())
        feat.SetField("DN", 1)
        feat.SetGeometry(union_geom)
        out_lyr.CreateFeature(feat)
        feat = None
        out_ds = None

        _log("info", f"Cutline written: {output_cutline}")
        return output_cutline

    except Exception as e:
        _log("exception", f"Error creating cutline from grid: {e}")
        return None


def create_cutline_file(v_shift: bool,
                        grid_files: list[str],
                        cutline_path: str,
                        input_metadata: dict) -> Optional[str]:
    """
    Create a cutline file from the provided grid files if vertical shift is applied.
    Otherwise, return None.

    TODO: what if more than one NWLD grids are involved?
    """
    if v_shift and grid_files:
        # Use the first nwld grid file (or we could merge multiple grids)
        grid_file = grid_files[0]
        for gf in grid_files:
            if gf.lower().find("nwld") != -1:
                grid_file = gf
                break
        if not os.path.isabs(grid_file):
            vyper_grids = os.environ.get("VYPER_GRIDS", "")

            for base_dir in [vyper_grids]:
                potential_path = os.path.join(base_dir, grid_file)
                if os.path.exists(potential_path):
                    grid_file = potential_path
                    break

        if os.path.exists(grid_file):
            cutline_path = create_cutline_from_grid(grid_file, cutline_path,
                                                    input_extent=input_metadata["extent"],
                                                    input_wkt=input_metadata["wkt"],
                                                    )
            if cutline_path:
                logger.info(f"Using cutline from grid: {cutline_path} for grid: {grid_file}")
        else:
            logger.warning(f"Grid file not found: {grid_file}")
    else:
        cutline_path = None
    return cutline_path


def clip_raster_to_cutline(input_path: str, cutline_path: str, output_path: str) -> Optional[str]:
    """
    Clips a raster to a cutline while preserving original resolution and metadata.
    This creates a pre-masked file where all data points are guaranteed to be 
    within the grid's valid area for the subsequent transformation.
    """
    try:
        # Get original metadata to ensure we match resolution and nodata
        ds_in = gdal.Open(input_path, gdal.GA_ReadOnly)
        if ds_in is None:
            return None

        # Get input NoData and GeoTransform
        gt = ds_in.GetGeoTransform()
        x_res, y_res = abs(gt[1]), abs(gt[5])
        band = ds_in.GetRasterBand(1)
        nodata = band.GetNoDataValue() if band.GetNoDataValue() is not None else -9999.0

        # Use gdal.Warp to clip. We DO NOT change the CRS yet.
        # This keeps the math local and simple.
        warp_options = gdal.WarpOptions(
            format="GTiff",
            cutlineDSName=cutline_path,
            cropToCutline=True,  # Shrink the file extent to the grid overlap
            dstNodata=nodata,
            xRes=x_res,
            yRes=y_res,
            resampleAlg="near",
            creationOptions=["COMPRESS=DEFLATE", "TILED=YES"]
        )

        # Run the clip
        ds_out = gdal.Warp(output_path, ds_in, options=warp_options)

        if ds_out:
            # Transfer metadata tags (like Vyperdatum_Metadata)
            ds_out.SetMetadata(ds_in.GetMetadata())
            ds_out.FlushCache()
            ds_out = None
            ds_in = None
            return output_path

        return None
    except Exception as e:
        logging.error(f"Error in clip_raster_to_cutline: {e}")
        return None



import os
import glob
from pathlib import Path
from vyperdatum.transformer import Transformer
from vyperdatum.utils.raster_utils import raster_metadata, update_raster_wkt
from vyperdatum.utils.vdatum_rest_utils import vdatum_cross_validate
import pyproj as pp
from osgeo import gdal
import re


def get_tiff_files(parent_dir: str, extention: str) -> list:
    tiff_files = []
    for (dirpath, dirnames, filenames) in os.walk(parent_dir):
        for filename in filenames:
            if filename.endswith(extention):
                tiff_files.append(os.sep.join([dirpath, filename]))
    return tiff_files


def transform_with_vyperdatum_navd88(input_file, zone):
    if zone == "16N":
        crs_from = "EPSG:26916+NOAA:89"
        crs_to = "EPSG:26916+EPSG:5703"
    elif zone == "15N":
        crs_from = "EPSG:26915+NOAA:89"
        crs_to = "EPSG:26915+EPSG:5703"
    else:
        raise ValueError(f"Invalid zone: {zone}")

    tf = Transformer(crs_from=crs_from,
                     crs_to=crs_to,
                    #  steps=steps
                     )
    output_file = input_file.replace("Original", "Manual")
    tf.transform_raster(input_file=input_file,
                        output_file=output_file,
                        overview=False,
                        pre_post_checks=True,
                        vdatum_check=False
                        )

    # passed, cross_df = vdatum_cross_validate(s_wkt=pp.CRS(crs_from).to_wkt(),
    #                                          t_wkt=pp.CRS(crs_to).to_wkt(),
    #                                          n_sample=20,
    #                                          s_raster_metadata=raster_metadata(input_file),
    #                                          t_raster_metadata=raster_metadata(output_file),
    #                                          s_point_samples=None,
    #                                          t_point_samples=None,
    #                                          tolerance=0.3,
    #                                          raster_sampling_band=1,
    #                                          region="contiguous",
    #                                          pivot_h_crs="EPSG:6318",
    #                                          s_h_frame="NAD83_2011",
    #                                          s_v_frame="NAVD88",
    #                                          s_h_zone=None,
    #                                          t_h_frame="NAD83_2011",
    #                                          t_v_frame="LWD_IGLD85",
    #                                          t_h_zone=None
    #                                          )
    # cross_df.to_csv(Path(output_file).parent.absolute()/Path(f"{os.path.split(input_file)[1]}_vdatum_check.csv"), index=False)    
    return


if __name__ == "__main__":
    files = glob.glob(r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\MSP\Original\**\*.tif", recursive=True)

    for i, input_file in enumerate(files[1:2]):
        try:
            print(f"{i+1}/{len(files)}: {input_file}")
            zone = None
            if input_file.find(r"\16N") != -1:
                zone = "16N"
            elif input_file.find(r"\15N") != -1:
                zone = "15N"
            # meta = raster_metadata(input_file)
            # print(f"raster horizontal CRS: {meta['h_authcode']}\nraster vertical CRS: {meta['v_authcode']}")
            transform_with_vyperdatum_navd88(input_file, zone)

            print(f'\n{"*"*50} {i+1}/{len(files)} Completed {"*"*50}\n')
        except Exception as e:
            print(f"Error transforming {input_file}:\n{e}")





import pathlib
from osgeo import gdal, ogr
import numpy as np
import pyproj as pp
from vyperdatum.transformer import Transformer


def vector_wkt(vector_file: str):
    ds = gdal.OpenEx(vector_file)
    driver = ogr.GetDriverByName(ds.GetDriver().ShortName)
    ds = None
    dataSource = driver.Open(vector_file, 0)
    layer = dataSource.GetLayer()
    wkt = layer.GetSpatialRef().ExportToWkt()
    driver, layer = None, None
    return wkt


def test_transform_vector_Hudson():
    input_file = r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\vector\Modeling_Tile_Scheme_20240320.gpkg"
    transformed_file = (str(pathlib.Path(input_file).with_stem("_transformed_"
                                                               + pathlib.Path(input_file).stem)))



    T = Transformer(crs_from="EPSG:6318",
                    crs_to="EPSG:6318+NOAA:98",
                    )                                                           
    # T = Transformer(crs_from=vector_wkt(input_file),
    #                 crs_to=f"{pp.CRS(vector_wkt(input_file)).to_2d().to_string()}+NOAA:98"
    #                 )
    assert T.transform_vector(input_file=input_file, output_file=transformed_file), "Hudson river vector transformation failed"




test_transform_vector_Hudson()


import os
import glob
import pathlib
import pandas as pd
from vyperdatum.transformer import Transformer
import geopandas as gpd
from shapely.geometry import Point


def get_skiprows(fname: str):
    ln = 0
    with open(fname, "r") as f:
        for line in f:
            ln += 1
            if line.startswith("append_to_abstract=="):
                return ln
    return None


if __name__ == "__main__":
    files = glob.glob(r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\point\USACE\PBA\UTM9N\Original\**\*.XYZ", recursive=True)

    # steps = [
    #         {"crs_from": "ESRI:102445", "crs_to": "EPSG:6319"},
    #         {"crs_from": "EPSG:6319", "crs_to": "EPSG:7912"},
    #         {"crs_from": "EPSG:7912", "crs_to": "EPSG:9989"},
    #         {"crs_from": "EPSG:9990+NOAA:98", "crs_to": "EPSG:9990+NOAA:5537"},
    #         {"crs_from": "EPSG:9989", "crs_to": "EPSG:7912"},
    #         {"crs_from": "EPSG:7912", "crs_to": "EPSG:6319"},
    #         {"crs_from": "EPSG:6319", "crs_to": "EPSG:6337"},
    #         # {"crs_from": "EPSG:6319", "crs_to": "ESRI:102445"},
    #         ]
    steps = [
            # {"crs_from": "ESRI:102445", "crs_to": "EPSG:6337"}  # UTM Zone 8N
            {"crs_from": "ESRI:102445", "crs_to": "EPSG:6338"}  # UTM Zone 9N
            ]
    

    for i, input_file in enumerate(files[:]):
        print(f"{i+1}/{len(files)}: {input_file}")
        df = pd.read_csv(input_file, sep=",", names=["x", "y", "z"], skiprows=get_skiprows(input_file))
        x, y, z = df.x.values, df.y.values, df.z.values
        tf = Transformer(crs_from=steps[0]["crs_from"],
                         crs_to=steps[-1]["crs_to"],
                         steps=steps
                         )
        success, xt, yt, zt = tf.transform_points(x, y, z,
                                                always_xy=True,
                                                allow_ballpark=False,
                                                only_best=True,
                                                vdatum_check=False)
        output_file = input_file.replace("Original", "Manual")
        pathlib.Path(os.path.split(output_file)[0]).mkdir(parents=True, exist_ok=True)
        csv_fname = output_file+".csv"
        tdf = pd.DataFrame({"x": xt, "y": yt, "z": zt})
        # tdf.to_csv(csv_fname, index=False)
        tdf['geometry'] = tdf.apply(lambda row: Point(row['x'], row['y'], row['z']), axis=1)
        gdf = gpd.GeoDataFrame(tdf, geometry='geometry', crs=f'{steps[-1]["crs_to"]}')
        gdf.to_file(f'{output_file.replace("XYZ", "gpkg")}', driver="GPKG")



from vyperdatum.transformer import Transformer


input_file = r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\geoparquet_test\South_Locust_Point_&_Masonville_&_Ferry_Bar_&_Fairfield_2023_Elevation_MLLW_3ftx3ft.parquet"
output_file = r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\geoparquet_test\output.parquet"

crs_from = "EPSG:6347+NOAA:98"
crs_to = "EPSG:6347+EPSG:5703"
tf = Transformer(crs_from=crs_from, crs_to=crs_to)
tf.transform(input_file=input_file,
             output_file=output_file,
             pre_post_checks=True,
             vdatum_check=True
             )


import numpy as np
import pandas as pd
import geopandas as gpd
from shapely.geometry import Point
from pyproj import Transformer
import pyproj as pp

def generate_random_utm_points(utm_zone, n_points=100):
    """
    Generate random geographic points in a given UTM zone.

    Parameters:
        utm_zone (int): UTM zone number (1 to 60).
        hemisphere (str): 'N' for Northern or 'S' for Southern Hemisphere.
        n_points (int): Number of points to generate.

    Returns:
        GeoDataFrame: with geometry, UTM, and lat/lon coordinates.
    """
    if not 1 <= utm_zone <= 60:
        raise ValueError("UTM zone must be between 1 and 60.")
    hemisphere = "N"


    # Geographic bounds of UTM zone
    lon_min = (utm_zone - 1) * 6 - 180
    lon_max = lon_min + 6
    lat_min = 0 if hemisphere.upper() == 'N' else -80
    lat_max = 84 if hemisphere.upper() == 'N' else 0

    # Generate random lat/lon within bounds
    lats = np.random.uniform(lat_min, lat_max, n_points)
    lons = np.random.uniform(lon_min, lon_max, n_points)



    crs_utm_83 = f"EPSG:{26900 + utm_zone}"
    crs_geodetic_83 = ":".join(pp.CRS(crs_utm_83).geodetic_crs.to_authority())

    crs_utm_2011 = f"EPSG:{6329 + utm_zone}"
    crs_geodetic_2011 = ":".join(pp.CRS(crs_utm_2011).geodetic_crs.to_authority())


    # Transform to UTM
    transformer_83 = Transformer.from_crs(crs_geodetic_83, crs_utm_83, always_xy=True)
    eastings_83, northings_83 = transformer_83.transform(lons, lats)

    transformer_2011 = Transformer.from_crs(crs_geodetic_2011, crs_utm_2011, always_xy=True)
    eastings_2011, northings_2011 = transformer_2011.transform(lons, lats)

    # Convert zipped pairs to NumPy arrays
    coords_83 = np.column_stack((eastings_83, northings_83))
    coords_2011 = np.column_stack((eastings_2011, northings_2011))

    # Compute deviation (Euclidean norm of differences)
    deviation = np.linalg.norm(coords_2011 - coords_83, axis=1)

    # Create GeoDataFrame
    df = pd.DataFrame({
        'Zone': [utm_zone] * n_points,
        'NAD83': [crs_geodetic_83] * n_points,
        'UTM_83': [crs_utm_83] * n_points,
        'NAD83_2011': [crs_geodetic_2011] * n_points,
        'UTM_83_2011': [crs_utm_2011] * n_points,
        'Longitude': lons,
        'Latitude': lats,
        'Easting_83': eastings_83,
        'Northing_83': northings_83,
        'Easting_2011': eastings_2011,
        'Northing_2011': northings_2011,
        'Deviation': deviation
    })

    return df


if __name__ == "__main__":
    utm_zone = 17  # Example UTM zone
    n_points = 100  # Number of points to generate
    df = generate_random_utm_points(utm_zone, n_points)
    print(df)


import logging
from typing import Union, Optional
from colorama import Fore, Style
from itertools import islice
import pyproj as pp
import networkx as nx
from tqdm import tqdm
from vyperdatum.db import DB
from vyperdatum.utils.crs_utils import validate_transform_steps, auth_code

logger = logging.getLogger("root_logger")


class Pipeline():
    """
    Requirements:
    source and target CRS must have explicit code in database
    (not any custom-built compound CRS will work).

    Assumes:
    Both source and target crs's horizontal crs are 3d (note the `.to_3d()` transforms)
    """
    def __init__(self,
                 crs_from: Union[pp.CRS, int, str],
                 crs_to: Union[pp.CRS, int, str],
                 filters: Optional[list[str]] = None
                 ) -> None:
        """

        Parameters
        ----------
        crs_from: pyproj.crs.CRS or input used to create one
            Projection of input data.
        crs_to: pyproj.crs.CRS or input used to create one
            Projection of output data.
        filters: list[str], optional, default=None
            WHERE clause conditions to filter the transformers in the db's
            `coordinate_operation_view`. If more than one condition is passed,
            logical `AND` is applied. Example: `filters=["auth_name='NOAA'"]`.
            Only applicable to the graph-based algorithm.

        Raises
        -------
        ValueError:
            If either code or authority of `crs_from`/`crs_to` (or their sub_crs)
            can not be determined. A compound CRS must have explicit auth/code in the database.

        """
        self.steps, self.head, self.tail = [], [], []
        self.G, self.nodes = None, []
        self.filters = filters
        self.WEIGHT = 1
        if not isinstance(crs_from, pp.CRS):
            crs_from = pp.CRS(crs_from)
        if not isinstance(crs_to, pp.CRS):
            crs_to = pp.CRS(crs_to)
        self.crs_from = crs_from
        self.crs_to = crs_to

        if self.crs_from.is_projected:
            self.head = [auth_code(self.crs_from, raise_no_auth=False)]
            if self.crs_from.is_compound:
                h = auth_code(pp.CRS(self.crs_from.sub_crs_list[0]).geodetic_crs)
                v = auth_code(pp.CRS(self.crs_from.sub_crs_list[1]))
                self.crs_from = pp.CRS(f"{h}+{v}")
            else:
                if self.crs_from.geodetic_crs.to_3d().to_authority():
                    self.crs_from = pp.CRS(auth_code(self.crs_from.geodetic_crs.to_3d()))
                else:
                    self.crs_from = pp.CRS(auth_code(self.crs_from.geodetic_crs))

        if self.crs_to.is_projected:
            self.tail = [auth_code(self.crs_to, raise_no_auth=False)]
            if self.crs_to.is_compound:
                h = auth_code(pp.CRS(self.crs_to.sub_crs_list[0]).geodetic_crs)
                v = auth_code(pp.CRS(self.crs_to.sub_crs_list[1]))
                self.crs_to = pp.CRS(f"{h}+{v}")
            else:
                if self.crs_to.geodetic_crs.to_3d().to_authority():
                    self.crs_to = pp.CRS(auth_code(self.crs_to.geodetic_crs.to_3d()))
                else:
                    self.crs_to = pp.CRS(auth_code(self.crs_to.geodetic_crs))

        self.s_h, self.s_v = self.split_crs(self.crs_from)
        self.t_h, self.t_v = self.split_crs(self.crs_to)
        self.crs_from = auth_code(self.crs_from)
        self.crs_to = auth_code(self.crs_to)
        self._compatible_h_for_v()
        return

    def _compatible_h_for_v(self) -> None:
        """
        Check if the horizontal CRS used for the compound output CRS is compatible with
        the vertical CRS. If not, replace with a valid horizontal CRS and prepend the original
        compound CRS to the tail list. The input horizontal and vertical CRSs are expected to
        be compatible already.

        Raises
        -------
        ValueError:
            When there is no matching horizontal CRS for the vertical component of the
            output CRS.
        """
        if pp.CRS(self.crs_to).is_compound:
            _, s_v = self.split_crs(pp.CRS(self.crs_from))
            h, v = self.split_crs(pp.CRS(self.crs_to))
            if s_v is None or v is None or s_v.upper() == v.upper():
                return
            va, vc = v.split(":")
            df = DB().query("select concat(horiz_crs_auth_name, ':', horiz_crs_code) hac from"
                            f" compound_crs where vertical_crs_auth_name='{va}'"
                            f" and vertical_crs_code='{vc}'", dataframe=True
                            )
            if h not in df["hac"].values:
                s_h = ":".join(pp.CRS(self.crs_from).geodetic_crs.to_2d().to_authority())
                if len(df) == 1:
                    compatible_h = df['hac'].values[0]
                elif s_h in df["hac"].values:
                    compatible_h = s_h
                else:
                    raise ValueError(f"{Fore.RED}Unable to find a matching output horizontal CRS"
                                     f"  for {v}.The followings are potential compatible"
                                     f" horizontal CRSs for {v}:\n{df['hac']}")
                self.tail = [self.crs_to] + self.tail
                self.crs_to = auth_code(f"{compatible_h}+{v}")
                self.t_h, self.t_v = compatible_h, v
        return

    @staticmethod
    def split_crs(crs: pp.CRS) -> tuple[str, Optional[str]]:
        """
        Return CRS string representation in form of code:authority

        Raises
        -------
        ValueError:
            If vertical-only CRS is received.

        Returns
        --------
        tuple:
            crs components in form of code:authority
        """
        if crs.is_vertical and not crs.is_compound:
            raise ValueError(f"The pipeline doesn't accept vertical-only crs:\n{crs}")
        h = ":".join(crs.geodetic_crs.to_authority(min_confidence=100))
        v = None
        if crs.is_compound:
            v = ":".join(pp.CRS(crs.sub_crs_list[1]).to_authority(min_confidence=100))
        return h, v

    @staticmethod
    def join_crs(h_crs: str, v_crs: str) -> str:
        """
        Return CRS string representation in form of code:authority

        Raises
        -------
        ValueError:
            If horizontal CRS is None; or when the CRS is invalid.

        Returns
        --------
        str:
            Concatenate the horizontal and vertical (if exists) CRS components using a `+` sign.
        """
        if not h_crs:
            raise ValueError("The horizontal CRS must be set.")
        try:
            _crs = f"{h_crs}{'+' + v_crs if v_crs else ''}"
            pp.CRS(_crs)
        except Exception as e:
            logger.exception(f"Unable to build crs: {_crs}\n{e}")
        return _crs

    def linear_steps(self) -> Optional[list[str]]:
        """
        Return a list of CRSs representing the transformation steps from `crs_from` to `crs_to`.
        If the generated pipeline can't be validated by PROJ, return `None`.
        """
        self.steps = []
        if f"{self.s_h}+{self.s_v}" == f"{self.t_h}+{self.t_v}":
            return []
        self.steps.append(self.join_crs(self.s_h, self.s_v))
        while (self.join_crs(*self.split_crs(pp.CRS(self.steps[-1])))
               !=
               self.join_crs(self.t_h, self.t_v)
               ):
            _, cur_v = self.split_crs(pp.CRS(self.steps[-1]))
            cur_h = ":".join(pp.CRS(self.steps[-1]).geodetic_crs.to_2d().to_authority())
            t_h = ":".join(pp.CRS(self.crs_to).geodetic_crs.to_2d().to_authority())
            if cur_h == t_h:
                self.steps.append(self.join_crs(t_h, self.t_v))
            else:
                if cur_h in ("EPSG:6318", "EPSG:4326") and t_h in ("EPSG:6318", "EPSG:4326"):
                    self.steps.append(self.join_crs("EPSG:9755", cur_v))
                if (cur_v is None and
                    pp.CRS(t_h).geodetic_crs.to_3d() and
                    (pp.CRS(self.crs_to).is_compound or len(pp.CRS(self.crs_to).axis_info) > 2)
                    ):  # noqa: E125
                    self.steps.append(":".join(pp.CRS(t_h).geodetic_crs.to_3d().to_authority()))
                else:
                    self.steps.append(self.join_crs(t_h, cur_v))
        temp_steps = self.head + self.steps + self.tail
        if not validate_transform_steps(temp_steps):
            logger.error(f"{Fore.RED}The following transformation steps generated using the"
                         f" linear algorithm: {temp_steps}, but PROJ cannot instantiate"
                         " valid transformer objects using them.")
            print(Style.RESET_ALL)
            return None
        self.steps = temp_steps
        return self.steps

    def build_graph(self):
        """
        Construct a graph model of the transformers defined in the proj.db.
        """
        self.G = nx.DiGraph()
        where_clause = "where deprecated=0"
        if self.filters:
            where_clause += " and " + " and ".join(self.filters)
        sql = f"select * from coordinate_operation_view {where_clause}"
        df = DB().query(sql, dataframe=True)
        logger.info("Building the transformer graph ...")
        for _, row in tqdm(df.iterrows(), total=df.shape[0]):
            self.G.add_edge(f"{row['source_crs_auth_name']}:{row['source_crs_code']}",
                            f"{row['target_crs_auth_name']}:{row['target_crs_code']}",
                            weight=self.WEIGHT
                            )
        self.add_extensions()
        return

    def _add_edge(self, node1, node2, weight, bidirectional=False):
        if not self.G.has_edge(node1, node2):
            self.G.add_edge(node1, node2, weight=weight)
        if bidirectional and not self.G.has_edge(node2, node1):
            self.G.add_edge(node2, node1, weight=weight)
        return

    def _remove_edge(self, node1, node2, bidirectional=False):
        if self.G.has_edge(node1, node2):
            self.G.remove_edge(node1, node2)
        if bidirectional and self.G.has_edge(node2, node1):
            self.G.remove_edge(node2, node1)
        return

    def add_extensions(self):
        """
        Add known valid transformation edges that are not explicitly listed in the database.
        Remove some of the known incorrect transformers in the database.
        """
        self._add_edge("EPSG:6319", "EPSG:7912", weight=self.WEIGHT, bidirectional=True)

        self._remove_edge("EPSG:4326", "EPSG:6318", bidirectional=True)
        self._add_edge("EPSG:4326", "EPSG:9755", weight=self.WEIGHT/10, bidirectional=True)
        self._add_edge("EPSG:9755", "EPSG:6318", weight=self.WEIGHT/10, bidirectional=True)

        self._add_edge("EPSG:4979", "EPSG:4269", weight=self.WEIGHT, bidirectional=True)
        self._add_edge("EPSG:4269", "EPSG:6319", weight=self.WEIGHT, bidirectional=True)
        self._add_edge("EPSG:4269", "EPSG:6318", weight=self.WEIGHT, bidirectional=True)
        return

    def add_node(self, new_node: str):
        nodes = list(self.G.nodes())
        if new_node in nodes:
            return
        logger.info(f"Adding node {new_node} to the transformers graph ...")
        new_crs = pp.CRS(new_node)
        for n in tqdm(nodes):
            try:
                t1 = pp.Transformer.from_crs(crs_from=new_crs,
                                             crs_to=pp.CRS(n),
                                             allow_ballpark=False,
                                             only_best=True
                                             )
                ps = t1.to_proj4()
                if ps and "+proj=noop" not in ps and "Error" not in ps:
                    self.G.add_edge(new_node, n, weight=self.WEIGHT)
            except:  # noqa: E722
                pass

            try:
                t2 = pp.Transformer.from_crs(crs_from=pp.CRS(n),
                                             crs_to=new_crs,
                                             allow_ballpark=False,
                                             only_best=True
                                             )
                ps = t2.to_proj4()
                if ps and "+proj=noop" not in ps and "Error" not in ps:
                    self.G.add_edge(n, new_node, weight=self.WEIGHT)
            except:  # noqa: E722
                pass
        return

    def k_graph_steps(self, k, weight="weight"):
        """
        Return `k` shortest (optimal) transformation steps between the source and
        target CRSs, ordered by `weight`.
        """
        if not self.G:
            self.build_graph()
            self.add_node(auth_code(self.crs_from))
            self.add_node(auth_code(self.crs_to))
        paths = list(islice(nx.shortest_simple_paths(self.G,
                                                     self.crs_from,
                                                     self.crs_to,
                                                     weight=weight), k))
        return [self.head + p + self.tail for p in paths]

    def graph_steps(self, weight="weight") -> Optional[list[str]]:
        """
        Return the optimal (shortest path) transformation steps between the source and target
        CRSs, measured by `weight`.
        """
        if not self.G:
            self.build_graph()
            self.add_node(auth_code(self.crs_from))
            self.add_node(auth_code(self.crs_to))
        temp_steps = self.head + nx.shortest_path(self.G,
                                                  self.crs_from,
                                                  self.crs_to,
                                                  weight=weight) + self.tail
        if not validate_transform_steps(temp_steps):
            logger.error(f"{Fore.RED}The following transformation steps generated using the"
                         f" graph-based optimal path algorithm: {temp_steps}, but PROJ cannot"
                         " instantiate valid transformer objects using them.")
            print(Style.RESET_ALL)
            return None
        self.steps = temp_steps
        return self.steps

    def transformation_steps(self, method: str = "linear") -> Optional[list[str]]:
        """
        Return transformation steps the source and target CRSs using an
        algorithm specified by `method`.

        Parameters
        -----------
        method: str, default 'linear'
            An algorithm using which the transformation steps are generated.
            Possible values: 'linear', 'graph'
        """
        method = method.lower().strip()
        if method == "linear":
            return self.linear_steps()
        elif method == "graph":
            return self.graph_steps()
        else:
            logger.error("Invalid method name to generate transformation steps.")
        return None


def nwld_ITRF2020_steps(h0: str, v0: Optional[str], h1: str, v1: Optional[str]):
    """
    Generate a general sequence of transformation steps from
    h0+v0 (crs_from) to h1+v1 (crs_to).
    """
    if v0 is None and v1 is None:
        return [{"crs_from": h0, "crs_to": h1, "v_shift": False}]

    steps = []
    if pp.CRS(h0).geodetic_crs.to_authority() == ("EPSG", "4326"):
        steps.append({"crs_from": h0, "crs_to": "EPSG:9755", "v_shift": False})
        steps.append({"crs_from": "EPSG:9755", "crs_to": "EPSG:6319", "v_shift": False})
    else:
        steps.append({"crs_from": h0, "crs_to": "EPSG:6319", "v_shift": False})
    steps.append({"crs_from": "EPSG:6319", "crs_to": "EPSG:7912", "v_shift": False})
    steps.append({"crs_from": "EPSG:7912", "crs_to": "EPSG:9989", "v_shift": False})

    if v0 is None:
        steps.append({"crs_from": "EPSG:9989", "crs_to": f"EPSG:9990+{v1}", "v_shift": True})
    elif v1 is None:
        steps.append({"crs_from": f"EPSG:9990+{v0}", "crs_to": "EPSG:9989", "v_shift": True})
    else:
        steps.append({"crs_from": f"EPSG:9990+{v0}", "crs_to": f"EPSG:9990+{v1}", "v_shift": True})

    steps.append({"crs_from": "EPSG:9990", "crs_to": "EPSG:9000", "v_shift": False})
    steps.append({"crs_from": "EPSG:9000", "crs_to": "EPSG:6318", "v_shift": False})
    if pp.CRS(h1).geodetic_crs.to_authority() == ("EPSG", "4326"):
        steps.append({"crs_from": "EPSG:6318", "crs_to": "EPSG:9755", "v_shift": False})
        steps.append({"crs_from": "EPSG:9755", "crs_to": h1, "v_shift": False})
    else:
        steps.append({"crs_from": "EPSG:6318", "crs_to": h1, "v_shift": False})
    return steps


def nwld_NAD832011_steps(h0: str, v0: Optional[str], h1: str, v1: Optional[str]):
    """
    Generate a general sequence of transformation steps from
    h0+v0 (crs_from) to h1+v1 (crs_to), assuming that v0 is
    compatible with NAD83 2011 (EPSG:6318).
    """
    if v0 is None and v1 is None:
        return [{"crs_from": h0, "crs_to": h1, "v_shift": False}]

    NAD83_geo = ["EPSG:6318", "EPSG:4269"]
    steps = []
    if pp.CRS(h0).geodetic_crs.to_authority() == ("EPSG", "4326"):
        steps.append({"crs_from": h0, "crs_to": "EPSG:9755", "v_shift": False})
        steps.append({"crs_from": "EPSG:9755", "crs_to": "EPSG:6318", "v_shift": False})
    elif h0 not in NAD83_geo:
        steps.append({"crs_from": h0, "crs_to": ":".join(pp.CRS(h0).geodetic_crs.to_authority()), "v_shift": False})

    # vertical shift step
    if (v0 is None or v1 is None) or v0.strip().lower() != v1.strip().lower():
        if v0 is None:
            steps.append({"crs_from": "EPSG:6319", "crs_to": f"EPSG:6318+{v1}", "v_shift": True})
        elif v1 is None:
            steps.append({"crs_from": f"EPSG:6318+{v0}", "crs_to": "EPSG:6319", "v_shift": True})
        else:
            steps.append({"crs_from": f"EPSG:6318+{v0}", "crs_to": f"EPSG:6318+{v1}", "v_shift": True})

    if pp.CRS(h1).geodetic_crs.to_authority() == ("EPSG", "4326"):
        steps.append({"crs_from": "EPSG:6318", "crs_to": "EPSG:9755", "v_shift": False})
        steps.append({"crs_from": "EPSG:9755", "crs_to": h1, "v_shift": False})
    elif h1 not in NAD83_geo:
        steps.append({"crs_from": "EPSG:6318", "crs_to": h1, "v_shift": False})
    return steps



from vyperdatum.transformer import Transformer
from osgeo import gdal



crs_from = "EPSG:6599+EPSG:5703"
crs_to = "EPSG:6339+NOAA:101"



input_file = r"C:\Users\mohammad.ashkezari\Desktop\clipped.tif"
output_file = r"C:\Users\mohammad.ashkezari\Desktop\clipped_trans.tif"
tf = Transformer(crs_from=crs_from,
                crs_to=crs_to,
                )

tf.transform(input_file=input_file,
            output_file=output_file,
            pre_post_checks=True,
            vdatum_check=False
            )


# ---------------------------------------------------------------
# crs_from = "EPSG:6346+NOAA:94"
# crs_to = "EPSG:6346+NOAA:93"



# # input_file = r"C:\Users\mohammad.ashkezari\Desktop\Fuse_Test\point\bag\Original\F00857\F00857_MB_50cm_LWD_1of1.bag"
# input_file = r"C:\Users\mohammad.ashkezari\Desktop\Fuse_Test\point\bag\Original\H11915\H11915_VB_5m_LWD_1of1.bag"
# tf = Transformer(crs_from=crs_from,
#                 crs_to=crs_to,
#                 )
# output_file = input_file.replace("Original", "Manual")
# tf.transform(input_file=input_file,
#             output_file=output_file,
#             pre_post_checks=True,
#             vdatum_check=False
#             )


# ---------------------------------------------------------------
# crs_from = "EPSG:6347"
# crs_to = "EPSG:6347+NOAA:98"

# input_file = r"C:\Users\mohammad.ashkezari\Desktop\Fuse_Test\raster\Original\nc\NC1901-TB-C_BLK-03_US4NC1AC_ellipsoidal_dem.tif"
# tf = Transformer(crs_from=crs_from,
#                  crs_to=crs_to,
#                  )
# output_file = input_file.replace("Original", "Manual")
# tf.transform_raster(input_file=input_file,
#                     output_file=output_file,
#                     overview=False,
#                     pre_post_checks=True,
#                     vdatum_check=True
#                     )




# ---------------------------------------------------------------

# PROJ9.4 produces an empty file (nodatavalues), PROJ6 fails and generates no file
# the target UTM zone is sets 17 which looks incorrect 


# crs_from = "EPSG:6347"
# crs_to = "EPSG:6346+NOAA:98"

# input_file = r"C:\Users\mohammad.ashkezari\Desktop\Fuse_Test\raster\Original\nc\NC1901-TB-C_BLK-03_US4NC1AC_ellipsoidal_dem.tif"
# tf = Transformer(crs_from=crs_from,
#                  crs_to=crs_to,
#                  )
# output_file = input_file.replace("Original", "Manual")
# tf.transform_raster(input_file=input_file,
#                     output_file=output_file,
#                     overview=False,
#                     pre_post_checks=True,
#                     vdatum_check=True
#                     )



# import pytest
# from vyperdatum.utils.raster_utils import raster_metadata
# from vyperdatum.utils.vdatum_rest_utils import (api_crs_aliases,
#                                                 vdatum_transform_point,
#                                                 vdatum_cross_validate_raster
#                                                 )


# @pytest.mark.parametrize("input_file, vdatum_h_crs, vdatum_v_crs", [
#     (r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\NVD\VA_input.tif", "ITRF2014", "LMSL"),
#     (r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\BlueTopo\BC25L26L\Modeling_BC25L26L_20230919.tiff", "NAD83_2011", "MLLW"),
#     (r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\BlueTopo\BC25L26L\_03_6318_5703_Modeling_BC25L26L_20230919.tiff", "NAD83_2011", "NAVD88"),
#     ])
# def test_api_crs_aliases(input_file: str, vdatum_h_crs: str, vdatum_v_crs: str):
#     meta = raster_metadata(input_file, verbose=False)
#     h_crs, v_crs = api_crs_aliases(meta["wkt"])
#     assert h_crs == vdatum_h_crs, (f"Horizontal CRS name ({h_crs}) didn't match"
#                                    f" Vdatum's alias ({vdatum_h_crs}) for {input_file}.")
#     assert v_crs == vdatum_v_crs, (f"Vertical CRS name ({v_crs}) didn't match"
#                                    f" Vdatum's alias ({vdatum_v_crs}) for {input_file}.")


# def test_vdatum_transform_point():
#     xx, yy, zz = -70.7, 43, 0
#     tx, ty, tz = -70.7, 43, -1.547
#     points, _ = vdatum_transform_point(s_x=xx, s_y=yy, s_z=zz, region="contiguous",
#                                        s_h_frame="NAD83_2011", s_v_frame="MLLW",
#                                        s_h_zone=None,
#                                        t_h_frame="NAD83_2011", t_v_frame="NAVD88",
#                                        t_h_zone=None
#                                        )
#     assert pytest.approx(points[0], abs=.01) == tx, "x coordinate should remain unchanged."
#     assert pytest.approx(points[1], abs=.01) == ty, "y coordinate should remain unchanged."
#     assert pytest.approx(points[2], abs=.2) == tz, (f"Vdatum transferred value ({points[2]})"
#                                                     f" doesn't match the expected value ({tz})")


# @pytest.mark.parametrize("s_file, t_file", [
#     # (r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\BlueTopo\BC25L26L\Modeling_BC25L26L_20230919.tiff",
#     #  r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\BlueTopo\BC25L26L\_03_6318_5703_Modeling_BC25L26L_20230919.tiff"
#     #  ),
#     (r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\PBB\Original\FL1703-TB-N\ellipsoid\FL1703-TB-N_US4FL2BJ_ellipsoidal_dem.tif",
#      r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\PBB\Manual\FL1703-TB-N\ellipsoid\FL1703-TB-N_US4FL2BJ_ellipsoidal_dem.tif"
#      )
#     ])
# def test_vdatum_cross_validate_raster(s_file: str, t_file: str):
#     passed, cross_df = vdatum_cross_validate_raster(s_file=s_file,
#                                                     t_file=t_file,
#                                                     n_sample=20,
#                                                     sampling_band=1,
#                                                     region=None,
#                                                     pivot_h_crs="EPSG:6318",
#                                                     s_h_frame=None,
#                                                     s_v_frame=None,
#                                                     s_h_zone=None,
#                                                     t_h_frame=None,
#                                                     t_v_frame=None,
#                                                     t_h_zone=None
#                                                     )
#     assert passed, f"Transformed raster inconsistent with Vdatum ({t_file})."



import os
import shutil
import time
from typing import TypedDict
import logging
import requests
from tqdm.auto import tqdm
import zipfile
from vyperdatum.enums import DATUM_DOI, ASSETS


logger = logging.getLogger("root_logger")


def datums_missing(datums_dir: str):
    """
    Rerturn True if the datum directory in the assets dir is empty or missing.
    """
    return not (os.path.isdir(datums_dir) and len(os.listdir(datums_dir)) > 0)


class DOI(TypedDict):
    url: str
    dir_name: str

def download_datums(doi: DOI):
    """
    Download datum files, unzip and copy them to the assets directory.
    """
    url = doi["url"]
    logger.info(f"Downloading datum and proj.db files from: {url}")
    resp = requests.get(url, stream=True)
    total = int(resp.headers.get("content-length", 0))
    zip_name = f"{ASSETS.DIR.value}/datums.zip"
    chunk_size = 1024
    with open(zip_name, "wb") as file, tqdm(
        desc=zip_name,
        total=total,
        unit="iB",
        unit_scale=True,
        unit_divisor=chunk_size,
    ) as bar:
        for data in resp.iter_content(chunk_size=chunk_size):
            size = file.write(data)
            bar.update(size)
    logger.info(f"Unzipping {zip_name}")
    with zipfile.ZipFile(zip_name, "r") as zip_ref:
        zip_ref.extractall(ASSETS.DIR.value)
    os.remove(zip_name)
    time.sleep(2)
    try:
        src_dir = f"{ASSETS.DIR.value}/{doi['dir_name']}"
        dst_dir = f"{ASSETS.DIR.value}/datums"
        shutil.move(src=src_dir, dst=dst_dir)
    except Exception as e:
        logger.exception(f"{e}\nUnable to rename a directory due to potential permission issue."
                         f" Please rename {src_dir} to {dst_dir} manually and run again.")
    return
